#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pandas.tseries.holiday import (
    Day,
    Easter,
    GoodFriday,
    Holiday,
    previous_friday,
)
from pytz import timezone

from .common_holidays import corpus_christi
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# Universal Confraternization (new years day)
ConfUniversal = Holiday(
    "Dia da Confraternizacao Universal",
    month=1,
    day=1,
)
# Sao Paulo city birthday
AniversarioSaoPaulo = Holiday(
    "Aniversario de Sao Paulo",
    month=1,
    day=25,
    # in 2022, BVMF was opened on this day
    end_date="2022-01-01",
)
# Carnival Monday
CarnavalSegunda = Holiday(
    "Carnaval Segunda", month=1, day=1, offset=[Easter(), Day(-48)]
)
# Carnival Tuesday
CarnavalTerca = Holiday("Carnaval Terca", month=1, day=1, offset=[Easter(), Day(-47)])
# Ash Wednesday (short day)
QuartaCinzas = Holiday(
    "Quarta Cinzas",
    month=1,
    day=1,
    offset=[Easter(), Day(-46)],
    start_date="2016-01-01",
)
# Good Friday
SextaPaixao = GoodFriday
# Feast of the Most Holy Body of Christ
CorpusChristi = corpus_christi()
# Tiradentes Memorial
Tiradentes = Holiday(
    "Tiradentes",
    month=4,
    day=21,
)
# Labor Day
DiaTrabalho = Holiday(
    "Dia Trabalho",
    month=5,
    day=1,
)
# Constitutionalist Revolution
Constitucionalista_prepandemic = Holiday(
    "Constitucionalista pre-pandemia",
    month=7,
    day=9,
    start_date="1998-01-01",
    end_date="2020-01-01",
)
Constitucionalista_pospandemic = Holiday(
    "Constitucionalista pos-pandemia",
    month=7,
    day=9,
    start_date="2021-01-01",
    end_date="2022-01-01",
)
# Independence Day
Independencia = Holiday(
    "Independencia",
    month=9,
    day=7,
)
# Our Lady of Aparecida
Aparecida = Holiday(
    "Nossa Senhora de Aparecida",
    month=10,
    day=12,
)
# All Souls' Day
Finados = Holiday(
    "Dia dos Finados",
    month=11,
    day=2,
)
# Proclamation of the Republic
ProclamacaoRepublica = Holiday(
    "Proclamacao da Republica",
    month=11,
    day=15,
)
# Day of Black Awareness
ConscienciaNegra = Holiday(
    "Dia da Consciencia Negra",
    month=11,
    day=20,
    start_date="2004-01-01",
    end_date="2022-01-01",
)
# Christmas Eve
VesperaNatal = Holiday(
    "Vespera Natal",
    month=12,
    day=24,
)
# Christmas
Natal = Holiday(
    "Natal",
    month=12,
    day=25,
)
# New Year's Eve
AnoNovo = Holiday(
    "Ano Novo",
    month=12,
    day=31,
    observance=previous_friday,
)


class BVMFExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for BM&F BOVESPA (BVMF).

    Open Time: 10:00 AM, Brazil/Sao Paulo
    Close Time:
    - Until 2019-11-01: 5:00 PM, Brazil/Sao Paulo
    - Starting from 2019-11-04: 6:00 PM, Brazil/Sao Paulo

    Regularly-Observed Holidays:
    - Universal Confraternization (New year's day, Jan 1)
    - Sao Paulo City Anniversary (Jan 25)
    - Carnaval Monday (48 days before Easter)
    - Carnaval Tuesday (47 days before Easter)
    - Ash Wednesday (short day / half day trading) (46 days before Easter)
    - Passion of the Christ (Good Friday, 2 days before Easter)
    - Corpus Christi (60 days after Easter)
    - Tiradentes (April 21)
    - Labor day (May 1)
    - Constitutionalist Revolution (July 9 after 1997)
    - Independence Day (September 7)
    - Our Lady of Aparecida Feast (October 12)
    - All Souls' Day (November 2)
    - Proclamation of the Republic (November 15)
    - Day of Black Awareness (November 20 after 2004)
    - Christmas (December 24 and 25)
    - Business day before New Year's Day
        - December 29 if NYE falls on a Sunday
        - December 30 if NYE falls on a Saturday
        - December 31 if NYE falls on Monday-Friday
    """

    name = "BVMF"

    tz = timezone("America/Sao_Paulo")

    regular_late_open = time(13)

    open_times = ((None, time(10)),)

    close_times = (
        (None, time(17, 0)),
        (pd.Timestamp("2019-11-04"), time(18, 0)),
    )

    @property
    def adhoc_holidays(self):
        # CopaDoMundo2014 Brazil hosted World Cup and played Croatia (and won 3-1!)
        return [pd.Timestamp("2014-06-12")]

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                ConfUniversal,
                AniversarioSaoPaulo,
                CarnavalSegunda,
                CarnavalTerca,
                SextaPaixao,
                CorpusChristi,
                Tiradentes,
                DiaTrabalho,
                Constitucionalista_prepandemic,
                Constitucionalista_pospandemic,
                Independencia,
                Aparecida,
                Finados,
                ProclamacaoRepublica,
                ConscienciaNegra,
                VesperaNatal,
                Natal,
                AnoNovo,
            ]
        )

    @property
    def special_opens(self):
        return [
            (
                self.regular_late_open,
                HolidayCalendar([QuartaCinzas]),
            ),
        ]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas.tseries.holiday import EasterMonday, GoodFriday, Holiday
from pytz import timezone

from .common_holidays import (
    ascension_day,
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import WEEKDAYS, HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

AscensionDay = ascension_day(end_date="2001")
WhitMonday = whit_monday(end_date="2002")

QueensDay = Holiday(
    "Queen's Day",
    month=4,
    day=30,
    end_date="2002",
)

LabourDay = european_labour_day()

ChristmasEve = christmas_eve(days_of_week=WEEKDAYS)
Christmas = christmas()
BoxingDay = boxing_day()

NewYearsEveBefore2002 = new_years_eve(end_date="2002")
NewYearsEveInOrAfter2002 = new_years_eve(
    start_date="2002",
    days_of_week=WEEKDAYS,
)


class XAMSExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Euronext Amsterdam exchange, and the primary calendar for
    the Netherlands.

    Open Time: 9:00 AM, CET (Central European Time)
    Close Time: 5:30 PM, CET (Central European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Good Friday
      - Easter Monday
      - Labour Day
      - Christmas Day
      - Boxing Day

    Early Closes:
      - Christmas Eve
      - New Year's Eve

    Other countries on the Euronext:
      - Belgium
      - France
      - Portugal
    """

    # Source: https://www.euronext.com/en/calendars-hours
    regular_early_close = time(14, 5)

    name = "XAMS"  # Euronext Amsterdam
    tz = timezone("Europe/Amsterdam")
    open_times = ((None, time(9)),)
    close_times = ((None, time(17, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                AscensionDay,
                WhitMonday,
                QueensDay,
                LabourDay,
                Christmas,
                BoxingDay,
                NewYearsEveBefore2002,
            ]
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar([ChristmasEve, NewYearsEveInOrAfter2002]),
            ),
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

from pytz import timezone

from .exchange_calendar import HolidayCalendar, ExchangeCalendar
from .xbkk_holidays import (
    ChakriMemorialDay,
    ChulalongkornDay,
    CoronationDay2016AndBefore,
    CoronationDay2019AndAfter,
    HalfYearHoliday,
    HMKingsBirthday,
    HMQueenMothersBirthday,
    HMQueensBirthday,
    KingBhumibolsBirthday,
    LabourDay,
    NewYearsDay,
    NewYearsEve,
    SongkranFestival1,
    SongkranFestival2,
    SongkranFestival3,
    ThailandConstitutionDay,
    ThePassingOfKingBhumibol,
    asanha_bucha,
    asanha_bucha_bridge_days,
    coronation_bridge_days,
    makha_bucha,
    misc_adhoc,
    new_years_bridge_days,
    queens_birthday_bridge_days,
    vesak,
    vesak_bridge_days,
)


class XBKKExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Stock Exchange of Thailand in Bangkok.

    Open Time: 10:00 AM, Indochina Time (ICT)
    Close Time: 4:30 PM, Indochina Time (ICT)

    Regularly-Observed Holidays:
      - New Year's Day
      - Makha Bucha
      - Chakri Memorial Day
      - Songkran Festival
      - Labour Day
      - Coronation Day
      - Vesak
      - Her Majesty The Queen's Birthday
      - Asanha Bucha
      - His Majesty The King's Birthday
      - Her Majesty The Queen Mother's Birthday
      - The Passing of King Bhumibol
      - Chulalongkorn Day
      - King Bhumibol's Birthday
      - Thailand Constitution Day
      - New Year's Eve

    Early Closes:
      - None
    """

    name = "XBKK"

    tz = timezone("Asia/Bangkok")

    open_times = ((None, time(10)),)

    close_times = ((None, time(16, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                ChakriMemorialDay,
                SongkranFestival1,
                SongkranFestival2,
                SongkranFestival3,
                LabourDay,
                CoronationDay2016AndBefore,
                CoronationDay2019AndAfter,
                HMQueensBirthday,
                HMKingsBirthday,
                HMQueenMothersBirthday,
                HalfYearHoliday,
                ThePassingOfKingBhumibol,
                ChulalongkornDay,
                KingBhumibolsBirthday,
                ThailandConstitutionDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                makha_bucha,
                vesak,
                asanha_bucha,
                new_years_bridge_days,
                asanha_bucha_bridge_days,
                queens_birthday_bridge_days,
                coronation_bridge_days,
                vesak_bridge_days,
                misc_adhoc,
            ),
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
import pytz

from .exchange_calendar import HolidayCalendar, ExchangeCalendar
from .xkls_holidays import (
    ChineseNewYear,
    ChineseNewYearDay2,
    ChineseNewYearsHalfDay,
    ChristmasDay,
    Deepavali,
    EidAlAdha,
    EidAlFitrDay1,
    EidAlFitrDay2,
    EidAlFitrHalfDay,
    FederalTerritoryDay,
    LabourDay,
    MalaysiaDay,
    MuhammadBirthday,
    Muharram,
    NationalDay,
    NewYearsDay,
    NuzulAlQuran,
    Thaipusam,
    WesakDay,
    misc_adhoc,
)


class XKLSExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Malaysia Stock Exchange (XKLS).

    Open Time: 9:00 AM, MYT
    Close Time: 5:00 PM, MYT

    Regularly-Observed Holidays:
    - New Year's Day
    - Thaipusam (Tamil Calendar)
    - Federal Territory Day (Feb 1)
    - Chinese New Year (2 days)
    - Labour Day (May 1)
    - Wesak Day (Chinese Lunar Calendar)
    - Nuzul Al'Quran (Islamic Lunar Calendar)
    - Eid al-Fitr (2 days)
    - Eid al-Adha
    - National Day (Aug 31)
    - Muharram (Lunar Calendar)
    - Malaysia Day (Sep 16)
    - Deepavali
    - Prophet Muhammad's Birthday
    - Christmas Day

    Early Closes:
    - Chinese New Year's Eve
    - Eid al-Fitr Eve

    TODO: XKLS takes a two hour lunch break from 12:30-2:30pm, but we are
    ignoring this for now.
    """

    name = "XKLS"

    tz = pytz.timezone("Asia/Kuala_Lumpur")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 00)),)

    regular_early_close = time(12, 30)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                FederalTerritoryDay,
                LabourDay,
                NationalDay,
                MalaysiaDay,
                ChristmasDay,
            ]
        )

    @property
    def adhoc_holidays(self):

        return list(
            chain(
                misc_adhoc,
                ChineseNewYear,
                ChineseNewYearDay2,
                NuzulAlQuran,
                EidAlFitrDay1,
                EidAlFitrDay2,
                EidAlAdha,
                Muharram,
                MuhammadBirthday,
                Deepavali,
                Thaipusam,
                WesakDay,
            )
        )

    @property
    def special_closes_adhoc(self):
        # Regular early closes on Chinese New Years Eve, Eid al-Fitr Eve
        early_close_days = pd.DatetimeIndex(
            set(ChineseNewYearsHalfDay + EidAlFitrHalfDay)
        )
        return [(self.regular_early_close, early_close_days)]



import pandas as pd
from dateutil.easter import EASTER_ORTHODOX, easter
from pandas.tseries.holiday import FR, DateOffset, Easter, Holiday
from pandas.tseries.offsets import Day

from .exchange_calendar import MONDAY, TUESDAY


def new_years_day(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "New Year's Day",
        month=1,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def new_years_eve(
    start_date=None,
    end_date=None,
    observance=None,
    days_of_week=None,
    offset=None,
):
    return Holiday(
        "New Year's Eve",
        month=12,
        day=31,
        start_date=start_date,
        end_date=end_date,
        offset=offset,
        observance=observance,
        days_of_week=days_of_week,
    )


def epiphany(
    start_date=None,
    end_date=None,
    observance=None,
    days_of_week=None,
    offset=None,
):
    return Holiday(
        "Epiphany",
        month=1,
        day=6,
        start_date=start_date,
        end_date=end_date,
        offset=offset,
        observance=observance,
        days_of_week=days_of_week,
    )


def anzac_day(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "Anzac Day",
        month=4,
        day=25,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def european_labour_day(
    start_date=None, end_date=None, observance=None, days_of_week=None
):
    return Holiday(
        "Labour Day",
        month=5,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


# Holy Wednesday, Maundy Thursday, Ascension Day, Whit Monday, and Corpus
# Christi do not take observance as a parameter because they depend on a
# particular offset, and offset and observance cannot both be passed to a
# Holiday.
def holy_wednesday(start_date=None, end_date=None, days_of_week=None):
    return Holiday(
        "Holy Wednesday",
        month=1,
        day=1,
        offset=[Easter(), -Day(4)],
        start_date=start_date,
        end_date=end_date,
        days_of_week=days_of_week,
    )


def maundy_thursday(start_date=None, end_date=None, days_of_week=None):
    return Holiday(
        "Maundy Thursday",
        month=1,
        day=1,
        offset=[Easter(), -Day(3)],
        start_date=start_date,
        end_date=end_date,
        days_of_week=days_of_week,
    )


def ascension_day(start_date=None, end_date=None):
    return Holiday(
        "Ascension Day",
        month=1,
        day=1,
        offset=[Easter(), Day(39)],
        start_date=start_date,
        end_date=end_date,
    )


def whit_monday(start_date=None, end_date=None):
    return Holiday(
        "Whit Monday",
        month=1,
        day=1,
        offset=[Easter(), Day(50)],
        start_date=start_date,
        end_date=end_date,
    )


def corpus_christi(start_date=None, end_date=None):
    return Holiday(
        "Corpus Christi",
        month=1,
        day=1,
        offset=[Easter(), Day(60)],
        start_date=start_date,
        end_date=end_date,
    )


def midsummer_eve(start_date=None, end_date=None):
    return Holiday(
        "Midsummer Eve",
        month=6,
        day=19,
        offset=DateOffset(weekday=FR(1)),
        start_date=start_date,
        end_date=end_date,
    )


def saint_peter_and_saint_paul_day(
    start_date=None, end_date=None, observance=None, days_of_week=None
):
    return Holiday(
        "Saint Peter and Saint Paul Day",
        month=6,
        day=29,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def assumption_day(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "Assumption Day",
        month=8,
        day=15,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def all_saints_day(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "All Saints Day",
        month=11,
        day=1,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def immaculate_conception(
    start_date=None, end_date=None, observance=None, days_of_week=None
):
    return Holiday(
        "Immaculate Conception",
        month=12,
        day=8,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def christmas_eve(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "Christmas Eve",
        month=12,
        day=24,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def christmas(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "Christmas",
        month=12,
        day=25,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def weekend_christmas(start_date=None, end_date=None, observance=None):
    """
    If christmas day is Saturday Monday 27th is a holiday
    If christmas day is sunday the Tuesday 27th is a holiday
    """
    return Holiday(
        "Weekend Christmas",
        month=12,
        day=27,
        days_of_week=(MONDAY, TUESDAY),
        start_date=start_date,
        end_date=end_date,
        observance=observance,
    )


def boxing_day(start_date=None, end_date=None, observance=None, days_of_week=None):
    return Holiday(
        "Boxing Day",
        month=12,
        day=26,
        start_date=start_date,
        end_date=end_date,
        observance=observance,
        days_of_week=days_of_week,
    )


def weekend_boxing_day(start_date=None, end_date=None, observance=None):
    """
    If boxing day is saturday then Monday 28th is a holiday
    If boxing day is sunday then Tuesday 28th is a holiday
    """
    return Holiday(
        "Weekend Boxing Day",
        month=12,
        day=28,
        days_of_week=(MONDAY, TUESDAY),
        start_date=start_date,
        end_date=end_date,
        observance=observance,
    )


# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Turkey's observance of Eid al-Fitr.  Other countries that
# observe Eid al-Fitr might use slightly different dates
eid_al_fitr_first_day = pd.to_datetime(
    [
        "1981-08-01",
        "1982-07-21",
        "1983-07-11",
        "1984-06-30",
        "1985-06-19",
        "1986-06-08",
        "1987-05-28",
        "1988-05-16",
        "1989-05-06",
        "1990-04-26",
        "1991-04-15",
        "1992-04-04",
        "1993-03-24",
        "1994-03-13",
        "1995-03-02",
        "1996-02-19",
        "1997-02-08",
        "1998-01-29",
        "1999-01-18",
        "2000-01-08",
        "2000-12-27",
        "2001-12-16",
        "2002-12-05",
        "2003-11-25",
        "2004-11-14",
        "2005-11-03",
        "2006-10-23",
        "2007-10-12",
        "2008-09-30",
        "2009-09-20",
        "2010-09-09",
        "2011-08-30",
        "2012-08-19",
        "2013-08-08",
        "2014-07-28",
        "2015-07-17",
        "2016-07-05",
        "2017-06-25",
        "2018-06-15",
        "2019-06-04",
        "2020-05-24",
        "2021-05-13",
        "2022-05-02",
        "2023-04-21",
        "2024-04-10",
        "2025-03-30",
        "2026-03-20",
        "2027-03-09",
        "2028-02-26",
        "2029-02-14",
        "2030-02-04",
        "2031-01-24",
        "2032-01-14",
        "2033-01-02",
        "2033-12-23",
        "2034-12-12",
        "2035-12-01",
        "2036-11-19",
        "2037-11-08",
        "2038-10-29",
        "2039-10-19",
        "2040-10-07",
        "2041-09-26",
        "2042-09-15",
        "2043-09-04",
        "2044-08-24",
        "2045-08-14",
        "2046-08-03",
        "2047-07-24",
        "2048-07-12",
        "2049-07-01",
    ]
)

# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Turkey's observance of Eid al-Adha.  Other countries that
# observe Eid al-Adha might use slightly different dates
eid_al_adha_first_day = pd.to_datetime(
    [
        "1981-10-08",
        "1982-09-27",
        "1983-09-17",
        "1984-09-05",
        "1985-08-26",
        "1986-08-15",
        "1987-08-04",
        "1988-07-23",
        "1989-07-13",
        "1990-07-02",
        "1991-06-22",
        "1992-06-11",
        "1993-05-31",
        "1994-05-20",
        "1995-05-09",
        "1996-04-27",
        "1997-04-17",
        "1998-04-07",
        "1999-03-27",
        "2000-03-16",
        "2001-03-05",
        "2002-02-22",
        "2003-02-10",
        "2004-02-01",
        "2005-01-20",
        "2006-01-09",
        "2006-12-31",
        "2007-12-20",
        "2008-12-08",
        "2009-11-27",
        "2010-11-16",
        "2011-11-06",
        "2012-10-25",
        "2013-10-15",
        "2014-10-04",
        "2015-09-24",
        "2016-09-12",
        "2017-09-01",
        "2018-08-21",
        "2019-08-11",
        "2020-07-31",
        "2021-07-20",
        "2022-07-09",
        "2023-06-28",
        "2024-06-16",
        "2025-06-06",
        "2026-05-27",
        "2027-05-16",
        "2028-05-05",
        "2029-04-24",
        "2030-04-13",
        "2031-04-02",
        "2032-03-22",
        "2033-03-11",
        "2034-03-01",
        "2035-02-18",
        "2036-02-07",
        "2037-01-26",
        "2038-01-16",
        "2039-01-05",
        "2039-12-26",
        "2040-12-14",
        "2041-12-04",
        "2042-11-23",
        "2043-11-12",
        "2044-10-31",
        "2045-10-21",
        "2046-10-10",
        "2047-09-30",
        "2048-09-19",
        "2049-09-08",
    ]
)


def orthodox_easter(start_date="1980", end_date="2022"):
    """
    This function gives a DatetimeIndex of Orthodox Easter dates from start_date to end_date
    """
    return pd.to_datetime(
        [
            easter(year, method=EASTER_ORTHODOX)
            for year in range(int(start_date), int(end_date))
        ]
    )



from datetime import time
from itertools import chain

from pandas import Timestamp
from pandas.tseries.holiday import (
    GoodFriday,
    USLaborDay,
    USPresidentsDay,
    USThanksgivingDay,
)
from pytz import timezone

from .exchange_calendar import ExchangeCalendar
from exchange_calendars.exchange_calendar import HolidayCalendar
from exchange_calendars.us_holidays import (
    Christmas,
    USIndependenceDay,
    USMartinLutherKingJrAfter1998,
    USMemorialDay,
    USNationalDaysofMourning,
    USNewYearsDay,
)


class IEPAExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for ICE US (IEPA).

    Open Time: 8pm, America/New_York
    Close Time: 6pm, America/New_York

    https://www.theice.com/publicdocs/futures_us/ICE_Futures_US_Regular_Trading_Hours.pdf # noqa
    """

    name = "IEPA"

    tz = timezone("America/New_York")

    open_times = ((None, time(20)),)

    close_times = ((None, time(18)),)

    @property
    def open_offset(self):
        return -1

    @property
    def special_closes(self):
        return [
            (
                time(13),
                HolidayCalendar(
                    [
                        USMartinLutherKingJrAfter1998,
                        USPresidentsDay,
                        USMemorialDay,
                        USIndependenceDay,
                        USLaborDay,
                        USThanksgivingDay,
                    ]
                ),
            )
        ]

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                USNationalDaysofMourning,
                # ICE was only closed on the first day of the Hurricane Sandy
                # closings (was not closed on 2012-10-30)
                [Timestamp("2012-10-29")],
            )
        )

    @property
    def regular_holidays(self):
        # https://www.theice.com/publicdocs/futures_us/exchange_notices/NewExNot2016Holidays.pdf # noqa
        return HolidayCalendar([USNewYearsDay, GoodFriday, Christmas])



import datetime
import pandas as pd

from pandas.tseries.offsets import Day
from korean_lunar_calendar import KoreanLunarCalendar
from pytz import timezone

from .holiday import Holiday
from .offsets import _is_normalized


def is_naive(dt):
    return dt.tzinfo is None


class KoreanHoliday(Holiday):

    _timezone = timezone("Asia/Seoul")
    _local_timezone = datetime.datetime.now().astimezone().tzinfo

    _computed_holidays = pd.Series([], index=pd.DatetimeIndex([]), dtype=object)
    _alternate_holidays_cache = pd.Series(
        pd.DatetimeIndex([]),
        index=pd.DatetimeIndex([]),
    )

    def _apply_rule(
        self, dates, apply_offset=True, apply_observance=True, register_holidays=True
    ):
        dates = super()._apply_rule(dates, apply_offset, apply_observance)
        if register_holidays:
            for date in dates:
                assert is_naive(date)
                assert _is_normalized(date)
                if date not in KoreanHoliday._computed_holidays.index:
                    KoreanHoliday._computed_holidays[date] = self
        return dates


def to_korean_datetime(dt):
    dt = pd.to_datetime(dt)
    if dt.tz is None:
        dt = dt.tz_localize(KoreanHoliday._local_timezone)
    dt = dt.tz_convert(KoreanHoliday._timezone)
    return dt


def is_saturday(dt):
    return dt.weekday() == 5


def is_sunday(dt):
    return dt.weekday() == 6


def is_holiday_saturday(dt):
    is_holiday_since = pd.Timestamp("1998-12-07")
    return dt >= is_holiday_since.tz_localize(dt.tzinfo) and is_saturday(dt)


def is_already_korean_holiday(dt):
    dt = pd.Timestamp(dt).tz_localize(None).normalize()
    return dt in KoreanHoliday._computed_holidays


original_alternative_holiday_since = pd.Timestamp(
    "2014-01-01"
)  # alternative holidays have been applied since year 2014
revised_alternative_holiday_since = pd.Timestamp(
    "2021-08-04"
)  # and expanded since 2021-08-04


def alternative_holiday_func(dt, is_already_holiday, since):
    if dt >= since:
        dt = pd.Timestamp(dt)
        dt_tzinfo = dt.tzinfo
        key_dt = dt.tz_localize(None).normalize()
        time_offset = dt.tz_localize(None) - key_dt
        if key_dt not in KoreanHoliday._alternate_holidays_cache.index:
            while is_already_holiday(dt):
                dt += Day(1)
            value_dt = dt.tz_localize(None).normalize()
            KoreanHoliday._alternate_holidays_cache[key_dt] = value_dt
        dt = KoreanHoliday._alternate_holidays_cache[key_dt]
        dt = dt + time_offset
        dt = dt.tz_localize(dt_tzinfo)
    return dt


def is_already_holiday(dt):
    return is_holiday_saturday(dt) or is_sunday(dt) or is_already_korean_holiday(dt)


def is_already_holiday_for_alternative_holiday_including_saturday(dt):
    # originally since 2014
    #  1. Children's Day
    # added since 2021-08-04
    #  2. Independence Movement Day
    #  3. National Liberation Day
    #  4. Korean National Foundation Day
    #  5. Hangul Proclamation Day
    return is_saturday(dt) or is_sunday(dt) or is_already_korean_holiday(dt)


def is_already_holiday_for_alternative_holiday_without_saturday(dt):
    # originally since 2014
    #  1. Seollal
    #  2. Chuseok
    return is_sunday(dt) or is_already_korean_holiday(dt)


def alternative_holiday_for_seollal_and_chuseok(dt):
    return alternative_holiday_func(
        dt,
        is_already_holiday_for_alternative_holiday_without_saturday,
        original_alternative_holiday_since,
    )


def alternative_holiday_for_childrens_day(dt):
    return alternative_holiday_func(
        dt,
        is_already_holiday_for_alternative_holiday_including_saturday,
        original_alternative_holiday_since,
    )


def alternative_holiday(dt):
    return alternative_holiday_func(
        dt,
        is_already_holiday_for_alternative_holiday_including_saturday,
        revised_alternative_holiday_since,
    )


def next_business_day(dt):
    while is_already_holiday(dt):
        dt += Day(1)
    return dt


def last_business_day(dt):
    while is_already_holiday(dt):
        dt -= Day(1)
    return dt


class KoreanSolarHoliday(KoreanHoliday):

    pass


def korean_lunar_to_solar(year, month, day, is_intercalation=False):
    calendar = KoreanLunarCalendar()
    is_valid = calendar.setLunarDate(year, month, day, is_intercalation)
    if not is_valid:
        raise ValueError(
            "Invalid date for lunar date: (year=%r, month=%r, day=%r, is_intercalation=%r)"
            % (year, month, day, is_intercalation)
        )
    return (calendar.solarYear, calendar.solarMonth, calendar.solarDay)


def korean_lunar_to_solar_datetime(dt, is_intercalation=False):
    year, month, day = korean_lunar_to_solar(
        dt.year, dt.month, dt.day, is_intercalation=is_intercalation
    )
    return dt.replace(year, month, day)


def korean_solar_to_lunar(year, month, day):
    calendar = KoreanLunarCalendar()
    is_valid = calendar.setSolarDate(year, month, day)
    if not is_valid:
        raise ValueError(
            "Invalid date for solar date: (year=%r, month=%r, day=%r)"
            % (year, month, day)
        )
    return (calendar.lunarYear, calendar.lunarMonth, calendar.lunarDay)


def korean_solar_to_lunar_datetime(dt, round_down: bool):
    """This method sets the year, month and day fields on a pd.Timestamp to the equivalent lunar ones.

    The problem is that not all lunar calendar dates can be stored in a Gregorian date.
    For example, 2022-03-31 (Gregorian) corresponds to 2022-02-29 (Korean lunar). The latter
    is not a valid Gregorian date. So, we introduce a rounding flag. If round_down is True,
    then this method will decrement the solar date until the lunar date happens to be a valid
    Gregorian one. If round_down is False, then the solar date will be incremented analogously."""
    while True:
        year, month, day = korean_solar_to_lunar(dt.year, dt.month, dt.day)
        try:
            datetime.date(year, month, day)
            return dt.replace(year, month, day)
        except ValueError:
            dt += datetime.timedelta(days=-1 if round_down else 1)


class KoreanLunarHoliday(KoreanHoliday):

    _max_solar_end_date = pd.to_datetime(
        str(KoreanLunarCalendar.KOREAN_SOLAR_MAX_VALUE), format="%Y%m%d"
    )
    _max_lunar_end_date = pd.to_datetime(
        str(KoreanLunarCalendar.KOREAN_LUNAR_MAX_VALUE), format="%Y%m%d"
    )

    def _reference_dates(self, start_date, end_date, strict=False):
        solar_start_date = start_date
        solar_end_date = end_date

        # Restrict date range to fall into supported range of korean_lunar_calendar library
        if solar_end_date > self._max_solar_end_date:
            if not strict and solar_start_date < self._max_solar_end_date:
                solar_end_date = self._max_solar_end_date
            else:
                raise ValueError(
                    "Cannot support date range after %r, but %r ~ %r given"
                    % (self._max_solar_end_date, start_date, end_date)
                )

        # Get lunar reference dates
        lunar_start_date = korean_solar_to_lunar_datetime(solar_start_date, round_down=True)
        lunar_end_date = korean_solar_to_lunar_datetime(solar_end_date, round_down=False)
        dates = super()._reference_dates(lunar_start_date, lunar_end_date)

        # Still restrict date range to fall into supported range of korean_lunar_calendar library
        dates = dates[dates <= self._max_lunar_end_date]

        # Convert lunar dates to solar dates
        dates = dates.to_series()
        dates = dates.map(korean_lunar_to_solar_datetime)
        dates = pd.DatetimeIndex(dates)

        return dates[(dates >= start_date) & (dates < end_date)]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

from pandas import DatetimeIndex
from pandas.tseries.holiday import GoodFriday, USLaborDay
import pytz

from .exchange_calendar import HolidayCalendar, ExchangeCalendar
from .us_holidays import (
    August45VictoryOverJapan,
    Christmas,
    ChristmasBefore1954,
    ChristmasEveBefore1993,
    ChristmasEveInOrAfter1993,
    ChristmasEvesAdhoc,
    DayAfterChristmasAdhoc,
    DayAfterIndependenceDayAdhoc,
    DayBeforeDecorationAdhoc,
    FirstLunarLandingClosing,
    FridayAfterIndependenceDayPre2013,
    HurricaneGloriaClosing,
    HurricaneSandyClosings,
    LincolnsBirthDayAdhoc,
    March33BankHoliday,
    MonTuesThursBeforeIndependenceDay,
    NewYorkCityBlackout77,
    November29BacklogRelief,
    PaperworkCrisis68,
    September11Closings,
    USBlackFridayBefore1993,
    USBlackFridayInOrAfter1993,
    USColumbusDayBefore1954,
    USElectionDay1848to1967,
    USElectionDay1968to1980,
    USIndependenceDay,
    USIndependenceDayBefore1954,
    USLincolnsBirthDayBefore1954,
    USMartinLutherKingJrAfter1998,
    USMemorialDay,
    USMemorialDay1964to1969,
    USMemorialDayBefore1964,
    USJuneteenth,
    USNationalDaysofMourning,
    USNewYearsDay,
    USPresidentsDay,
    USThanksgivingDay,
    USThanksgivingDay1939to1941,
    USThanksgivingDayBefore1939,
    USVeteransDay1934to1953,
    USWashingtonsBirthDay1964to1970,
    USWashingtonsBirthDayBefore1964,
    WeatherSnowClosing,
    WednesdayBeforeIndependenceDayPost2013,
)

# Useful resources for making changes to this file:
# http://www.nyse.com/pdfs/closings.pdf
# http://www.stevemorse.org/jcal/whendid.html


class XNYSExchangeCalendar(ExchangeCalendar):
    """
    Regularly-Observed Holidays:
    - New Years Day (observed on monday when Jan 1 is a Sunday)
    - Martin Luther King Jr. Day (3rd Monday in January, only after 1998)
    - Lincoln's Birthday (February 12th, only before 1954)
    - Washington's Birthday (February 22nd, before 1971 with rule change in
      1964)
    - Washington's Birthday (aka President's Day, 3rd Monday in February,
      after 1970)
    - Good Friday (two days before Easter Sunday)
    - Memorial Day (May 30th, before 1970, with rule change in 1964)
    - Memorial Day (last Monday in May, after 1970)
    - Juneteenth (June 16th, since 2022)
    - Independence Day (July 4th Sunday to Monday, before 1954)
    - Independence Day (observed on the nearest weekday to July 4th, after
      1953)
    - Election Day (First Tuesday starting on November 2nd, between 1848 and
      1967)
    - Election Day (Every four years, first Tuesday starting on November 2nd,
      between 1968 and 1980)
    - Veterans Day (November 11th, between 1934 and 1953)
    - Columbus Day (October 12th, before 1954)
    - Labor Day (first Monday in September)
    - Thanksgiving (last Thursday in November, before 1939)
    - Thanksgiving (second to last Thursday in November, between 1939 and 1941)
    - Thanksgiving (fourth Thursday in November, after 1941)
    - Christmas (December 25th, Sunday to Monday, before 1954)
    - Christmas (observed on nearest weekday to December 25, after 1953)

    NOTE: The NYSE does not observe the following US Federal Holidays:
    - Columbus Day (after 1953)
    - Veterans Day (after 1953)

    Regularly-Observed Early Closes:
    - July 3rd (Mondays, Tuesdays, and Thursdays, 1995 onward)
    - July 5th (Fridays, 1995 onward, except 2013)
    - Christmas Eve (except on Fridays, when the exchange is closed entirely)
    - Day After Thanksgiving (aka Black Friday, observed from 1992 onward)

    NOTE: Until 1993, the standard early close time for the NYSE was 2:00 PM.
    From 1993 onward, it has been 1:00 PM.

    Additional Irregularities:
    - Closed on 11/1/1929 and 11/29/1929 for backlog relief.
    - Closed between 3/6/1933 and 3/14/1933 due to bank holiday.
    - Closed on 8/15/1945 and 8/16/1945 following victory over Japan.
    - Closed on Christmas Eve in 1945 and 1946.
    - Closed on December 26th in 1958.
    - Closed the day before Memorial Day in 1961.
    - Closed on 11/25/1963 due to John F. Kennedy's death.
    - Closed for Lincoln's Birthday in 1968.
    - Closed a number of days between June 12th and  December 24th in 1968
      due to paperwork crisis.
    - Closed on 4/9/1968 due to Martin Luther King's death.
    - Closed the day after Independence Day in 1968.
    - Closed on 2/10/1969 due to weather (snow).
    - Closed on 3/31/1969 due to Dwight D. Eisenhower's death.
    - Closed on 7/21/1969 following the first lunar landing.
    - Closed on 12/28/1972 due to Harry S. Truman's death.
    - Closed on 1/25/1973 due to Lyndon B. Johnson's death.
    - Closed on 7/14/1977 due to New York City blackout.
    - Closed on 9/27/1985 due to Hurricane Gloria.
    - Closed on 4/27/1994 due to Richard Nixon's death.
    - Closed from 9/11/2001 to 9/16/2001 due to terrorist attacks in NYC.
    - Closed on 6/11/2004 due to Ronald Reagan's death.
    - Closed on 1/2/2007 due to Gerald Ford's death.
    - Closed on 10/29/2012 and 10/30/2012 due to Hurricane Sandy.
    - Closed on 12/5/2018 due to George H.W. Bush's death.
    - Closed at 1:00 PM on Wednesday, July 3rd, 2013
    - Closed at 1:00 PM on Friday, December 31, 1999
    - Closed at 1:00 PM on Friday, December 26, 1997
    - Closed at 1:00 PM on Friday, December 26, 2003

    NOTE: The exchange was **not** closed early on Friday December 26, 2008,
    nor was it closed on Friday December 26, 2014. The next Thursday Christmas
    will be in 2025.  If someone is still maintaining this code in 2025, then
    we've done alright...and we should check if it's a half day.
    """

    regular_early_close = time(13)

    name = "XNYS"

    tz = pytz.timezone("America/New_York")

    open_times = ((None, time(9, 30)),)

    close_times = ((None, time(16)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                USNewYearsDay,
                USMartinLutherKingJrAfter1998,
                USLincolnsBirthDayBefore1954,
                USWashingtonsBirthDayBefore1964,
                USWashingtonsBirthDay1964to1970,
                USPresidentsDay,
                GoodFriday,
                USMemorialDayBefore1964,
                USMemorialDay1964to1969,
                USMemorialDay,
                USJuneteenth,
                USIndependenceDayBefore1954,
                USIndependenceDay,
                USLaborDay,
                USThanksgivingDayBefore1939,
                USThanksgivingDay1939to1941,
                USThanksgivingDay,
                USElectionDay1848to1967,
                USElectionDay1968to1980,
                USVeteransDay1934to1953,
                USColumbusDayBefore1954,
                ChristmasBefore1954,
                Christmas,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                November29BacklogRelief,
                March33BankHoliday,
                August45VictoryOverJapan,
                ChristmasEvesAdhoc,
                DayAfterChristmasAdhoc,
                DayBeforeDecorationAdhoc,
                LincolnsBirthDayAdhoc,
                PaperworkCrisis68,
                DayAfterIndependenceDayAdhoc,
                WeatherSnowClosing,
                FirstLunarLandingClosing,
                September11Closings,
                NewYorkCityBlackout77,
                HurricaneGloriaClosing,
                HurricaneSandyClosings,
                USNationalDaysofMourning,
            )
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        MonTuesThursBeforeIndependenceDay,
                        FridayAfterIndependenceDayPre2013,
                        WednesdayBeforeIndependenceDayPost2013,
                        USBlackFridayInOrAfter1993,
                        ChristmasEveInOrAfter1993,
                    ]
                ),
            ),
            (
                time(14),
                HolidayCalendar(
                    [
                        ChristmasEveBefore1993,
                        USBlackFridayBefore1993,
                    ]
                ),
            ),
        ]

    @property
    def special_closes_adhoc(self):
        return [
            (
                self.regular_early_close,
                DatetimeIndex(
                    [
                        "1997-12-26",
                        "1999-12-31",
                        "2003-12-26",
                    ]
                ),
            )
        ]



from datetime import time, timedelta
from itertools import chain

from pandas.tseries.holiday import Holiday
from pytz import timezone

from .common_holidays import (
    christmas,
    european_labour_day,
    new_years_day,
    orthodox_easter,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

DayAfterNewYearsDay = Holiday("Day After New Year's Day", month=1, day=2)

RomanianPrincipalitiesUnificationDay = Holiday(
    "Romanian Principalities Unification Day", month=1, day=24
)

OrthodoxGoodFriday = orthodox_easter() - timedelta(2)

OrthodoxEasterMonday = orthodox_easter() + timedelta(1)

LabourDay = european_labour_day()


ChildrensDay = Holiday(
    "Children's Day",
    month=6,
    day=1,
)

OrthodoxPentecost = orthodox_easter() + timedelta(49)

DescentOfTheHolySpirit = orthodox_easter() + timedelta(50)

StMarysDay = Holiday(
    "St. Mary's day",
    month=8,
    day=15,
)

StAndrewsDay = Holiday(
    "St. Andrew's day",
    month=11,
    day=30,
)

NationalDay = Holiday(
    "National Day",
    month=12,
    day=1,
)

ChristmasDay = christmas()

SecondDayOfChristmas = Holiday(
    "Second Day of Christmas",
    month=12,
    day=26,
)


class XBSEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the BUCHAREST Stock Exchange (XBSE).

    Open Time: 10:00 AM, EET
    Close Time: 5:45 PM, EET

    Regularly-Observed Holidays:
      - New Year's Day
      - Day after New Year's Day
      - Romanian Principalities Unification Day
      - Orthodox Good Friday
      - Orthodox Easter
      - Labour Day
      - Orthodox Pentecost
      - Children's Day
      - Assumption of Virgin Mary
      - St Andrew's day
      - Christmas
      - Day after Christmas

    Early Closes:
      - None
    """

    name = "XBSE"

    tz = timezone("Europe/Bucharest")

    open_times = ((None, time(10)),)

    close_times = ((None, time(17, 45)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                DayAfterNewYearsDay,
                RomanianPrincipalitiesUnificationDay,
                LabourDay,
                ChildrensDay,
                StMarysDay,
                StAndrewsDay,
                NationalDay,
                ChristmasDay,
                SecondDayOfChristmas,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                OrthodoxGoodFriday,
                OrthodoxEasterMonday,
                OrthodoxPentecost,
                DescentOfTheHolySpirit,
            )
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pandas.tseries.holiday import MO, GoodFriday, Holiday
from pandas.tseries.offsets import DateOffset
from pytz import timezone

from .common_holidays import (
    christmas,
    european_labour_day,
    maundy_thursday,
    new_years_day,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

# Starting in 2007 Constitution Day is observed on the first Monday of Feb.
ConstitutionDay = Holiday("Constitution Day", month=2, day=5, end_date="2007")
ConstitutionDayObserved = Holiday(
    "Constitution Day Observed",
    month=2,
    day=1,
    offset=DateOffset(weekday=MO(1)),
    start_date="2007",
)

# Starting in 2007 Juarez's Birthay is observed on the third Monday of March.
JuarezBirthday = Holiday("Juarez's Birthday", month=3, day=21, end_date="2007")
JuarezBirthdayObserved = Holiday(
    "Juarez's Birthday Observed",
    month=3,
    day=1,
    offset=DateOffset(weekday=MO(3)),
    start_date="2007",
)

MaundyThursday = maundy_thursday()

LabourDay = european_labour_day()

IndependenceDay = Holiday("Independence Day", month=9, day=16)

AllSoulsDay = Holiday("All Souls' Day", month=11, day=2, start_date="2006")

# Starting in 2007 Revolution Day is observed on the third Monday of November.
RevolutionDay = Holiday("Revolution Day", month=11, day=20, end_date="2007")
RevolutionDayObserved = Holiday(
    "Revolution Day Observed",
    month=11,
    day=1,
    offset=DateOffset(weekday=MO(3)),
    start_date="2007",
)

BankingHoliday = Holiday("Banking Holiday", month=12, day=12)

Christmas = christmas()


class XMEXExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Mexican Stock Exchange (Bolsa Mexicana de Valores) in
    Mexico City.

    Open Time: 8:30 AM, CT (Central Time)
    Close Time: 3:00 PM, CT (Central Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Constitution Day
      - Juarez's Birthday
      - Maundy Thursday
      - Good Friday
      - Labour Day
      - Independence Day
      - All Souls' Day
      - Mexican Revolution
      - Banking Holiday
      - Christmas Day

    Holidays No Longer Observed:
      - None

    Early Closes:
      - None
    """

    name = "XMEX"
    tz = timezone("America/Mexico_City")

    open_times = ((None, time(8, 30)),)
    close_times = ((None, time(15)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                ConstitutionDay,
                ConstitutionDayObserved,
                JuarezBirthday,
                JuarezBirthdayObserved,
                MaundyThursday,
                GoodFriday,
                LabourDay,
                IndependenceDay,
                AllSoulsDay,
                RevolutionDay,
                RevolutionDayObserved,
                BankingHoliday,
                Christmas,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            # Bicentennial Celebration.
            pd.Timestamp("2010-09-17"),
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import Holiday, weekend_to_monday
from pytz import timezone

from .common_holidays import european_labour_day, new_years_day, new_years_eve
from .exchange_calendar import WEEKDAYS, HolidayCalendar, ExchangeCalendar


def new_years_eve_observance(holidays):
    # For some reason New Year's Eve was not a holiday these years.
    holidays = holidays[(holidays.year != 2008) & (holidays.year != 2009)]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def new_years_holiday_observance(holidays):
    # New Year's Holiday did not follow the next-non-holiday rule in 2016.
    holidays = holidays[(holidays.year != 2016)]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def orthodox_christmas_observance(holidays):
    # Orthodox Christmas did not follow the next-non-holiday rule these years.
    holidays = holidays[(holidays.year != 2012) & (holidays.year != 2017)]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


def defender_of_fatherland_observance(holidays):
    # Defender of the Fatherland Day did not follow the next-non-holiday rule
    # these years.
    holidays = holidays[
        (holidays.year != 2013) & (holidays.year != 2014) & (holidays.year != 2019)
    ]

    return pd.to_datetime([weekend_to_monday(day) for day in holidays])


NewYearsDay = new_years_day(observance=weekend_to_monday)
NewYearsHoliday = Holiday(
    "New Year's Holiday",
    month=1,
    day=2,
    observance=new_years_holiday_observance,
)
NewYearsHoliday2 = Holiday(
    "New Year's Holiday",
    month=1,
    day=3,
    start_date="2005",
    end_date="2012",
)
NewYearsHoliday3 = Holiday(
    "New Year's Holiday",
    month=1,
    day=4,
    start_date="2005",
    end_date="2012",
)
NewYearsHoliday4 = Holiday(
    "New Year's Holiday",
    month=1,
    day=5,
    start_date="2005",
    end_date="2012",
)
NewYearsHoliday5 = Holiday(
    "New Year's Holiday",
    month=1,
    day=6,
    start_date="2005",
    end_date="2012",
)

OrthodoxChristmas = Holiday(
    "Orthodox Christmas",
    month=1,
    day=7,
    observance=orthodox_christmas_observance,
)

DefenderOfTheFatherlandDay = Holiday(
    "Defender of the Fatherland Day",
    month=2,
    day=23,
    observance=defender_of_fatherland_observance,
)

WomensDay = Holiday(
    "Women's Day",
    month=3,
    day=8,
    observance=weekend_to_monday,
)

LabourDay = european_labour_day(observance=weekend_to_monday)

VictoryDay = Holiday(
    "Victory Day",
    month=5,
    day=9,
    observance=weekend_to_monday,
)

DayOfRussia = Holiday(
    "Day of Russia",
    month=6,
    day=12,
    observance=weekend_to_monday,
)

UnityDay = Holiday(
    "Unity Day",
    month=11,
    day=4,
    observance=weekend_to_monday,
    start_date="2005",
)

NewYearsEve = new_years_eve(
    observance=new_years_eve_observance,
    days_of_week=WEEKDAYS,
)


# Adhoc Holidays
# --------------

# All of the following "extensions" are bridge holidays, meaning they are
# either a Monday or Friday that is made into a holiday to fill in the gap
# between a Tuesday or Thursday holiday, respectively. Unfortunately having
# these bridge days is not consistently the rule, so they are treated as adhoc.
# This means that in the future there may be manual additions needed here.
new_years_extensions = [
    "2003-01-03",
    "2013-01-03",
    "2013-01-04",
    "2014-01-03",
]

orthodox_christmas_extensions = [
    "2003-01-06",
    "2005-01-10",
    "2008-01-08",
    "2009-01-08",
    "2009-01-09",
    "2010-01-08",
    "2011-01-10",
    "2016-01-08",
]

defender_of_the_fatherland_extensions = [
    "2006-02-24",
    "2010-02-22",
]

womens_day_extensions = [
    "2005-03-07",
    "2011-03-07",
    "2012-03-09",
]

labour_day_extensions = [
    "2002-05-02",
    "2002-05-03",
    "2003-05-02",
    "2004-05-04",
    "2007-04-30",
    "2008-05-02",
    "2012-04-30",
    "2015-05-04",
    "2016-05-03",
]

victory_day_extensions = [
    "2002-05-10",
    "2005-05-10",
    "2006-05-08",
    "2017-05-08",
]

day_of_russia_extensions = [
    "2003-06-13",
    "2007-06-11",
    "2008-06-13",
    "2012-06-11",
    "2014-06-13",
]

unity_day_extensions = [
    "2008-11-03",
    "2010-11-05",
]

misc_adhoc = [
    # Exchange Holidays.
    "2002-11-07",
    "2002-11-08",
    "2002-12-12",
    "2002-12-13",
    "2003-11-07",
    "2003-12-12",
    "2004-11-08",
    "2004-12-13",
    "2008-09-18",
    # Trading Suspended.
    "2008-10-10",
    "2008-10-27",
]


class XMOSExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Moscow Stock Exchange.

    Open Time: 10:00 AM, MSK (Moscow Standard Time)
    Close Time: 6:45 PM, MSK (Moscow Standard Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - New Year's Holiday
      - Orthodox Christmas Day
      - Defender of the Fatherland Day
      - Women's Day
      - Spring and Labour Day
      - Victory Day
      - Day of Russia
      - Unity Day
      - New Year's Eve

    Holidays No Longer Observed:
      - New Year's Holiday Week

    Early Closes:
      - None
    """

    name = "XMOS"

    tz = timezone("Europe/Moscow")

    open_times = ((None, time(10)),)

    close_times = ((None, time(18, 45)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NewYearsHoliday,
                NewYearsHoliday2,
                NewYearsHoliday3,
                NewYearsHoliday4,
                NewYearsHoliday5,
                OrthodoxChristmas,
                DefenderOfTheFatherlandDay,
                WomensDay,
                LabourDay,
                VictoryDay,
                DayOfRussia,
                UnityDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                new_years_extensions,
                orthodox_christmas_extensions,
                defender_of_the_fatherland_extensions,
                womens_day_extensions,
                labour_day_extensions,
                victory_day_extensions,
                day_of_russia_extensions,
                unity_day_extensions,
                misc_adhoc,
            ),
        )



from datetime import time
from itertools import chain

import pandas as pd
import pytz

from .exchange_calendar import HolidayCalendar, ExchangeCalendar
from .xtks_holidays import (
    AutumnalEquinoxes,
    ChildrensDay,
    CitizensHolidayGoldenWeek,
    CitizensHolidaySilverWeek,
    ComingOfAgeDay2000Onwards,
    ComingOfAgeDayThrough1999,
    ConstitutionMemorialDay,
    CultureDay,
    EmperorAkihitoBirthday,
    EmperorNaruhitoBirthday,
    EquityTradingSystemFailure,
    GreeneryDay2007Onwards,
    GreeneryDayThrough2006,
    HealthAndSportsDay2000OnwardsThrough2019,
    HealthAndSportsDay2020,
    HealthAndSportsDay2021,
    HealthAndSportsDay2022Onwards,
    HealthAndSportsDayThrough1999,
    LaborThanksgivingDay,
    MarineDay2003OnwardsThrough2019,
    MarineDay2020,
    MarineDay2021,
    MarineDay2022Onwards,
    MarineDayThrough2002,
    Misc2019Holidays,
    MountainDay2020,
    MountainDay2021,
    MountainDay2022Onwards,
    MountainDayThrough2019,
    NationalFoundationDay,
    NewYearsHolidayDec31,
    NewYearsHolidayJan1,
    NewYearsHolidayJan2,
    NewYearsHolidayJan3,
    RespectForTheAgedDay2003Onwards,
    RespectForTheAgedDayThrough2002,
    ShowaDay,
    VernalEquinoxes,
)


class XTKSExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Tokyo Stock Exchange

    First session: 9:00am - 11:30am
    Lunch Break: 11:30am - 12:30pm
    Second session: 12:30pm - 3:00pm

    Regularly-Observed Holidays (see xtks_holidays.py for more info):
    - New Year's Holidays (Dec. 31 - Jan. 3)
    - Coming of Age Day (second Monday of January)
    - National Foundation Day (Feb. 11)
    - Vernal Equinox (usually Mar 20-22)
    - Greenery Day (Apr. 29 2000-2006, May 4 2007-present)
    - Showa Day (Apr. 29 2007-present)
    - Constitution Memorial Day (May 3)
    - Citizen's Holiday (May 4 2000-2006, later replaced by Greenery Day)
    - Children's Day (May 5)
    - Marine Day (July 20 2000-2002, third Monday of July 2003-present)
    - Respect for the Aged Day (Sep. 15 2000-2002, third Monday
      of Sep. 2003-present)
    - Autumnal Equinox (usually Sept. 22-24)
    - Health-Sports Day (second Monday of October)
    - Culture Day (November 3)
    - Labor Thanksgiving Day (Nov. 23)
    - Emperor's Birthday (Dec. 23)

    Additional Irregularities:
    - Closed on October 1, 2020 due to equity trading system failure
    """

    name = "XTKS"

    tz = pytz.timezone("Asia/Tokyo")

    open_times = ((None, time(9)),)
    break_start_times = ((None, time(11, 30)),)
    break_end_times = ((None, time(12, 30)),)
    close_times = ((None, time(15)),)

    @classmethod
    def bound_start(cls) -> pd.Timestamp:
        # not tracking holiday info farther back than 1997
        return pd.Timestamp("1997-01-01")

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsHolidayDec31,
                NewYearsHolidayJan1,
                NewYearsHolidayJan2,
                NewYearsHolidayJan3,
                ComingOfAgeDayThrough1999,
                ComingOfAgeDay2000Onwards,
                NationalFoundationDay,
                GreeneryDayThrough2006,
                ShowaDay,
                ConstitutionMemorialDay,
                GreeneryDay2007Onwards,
                CitizensHolidayGoldenWeek,
                ChildrensDay,
                MarineDayThrough2002,
                MarineDay2003OnwardsThrough2019,
                MarineDay2020,
                MarineDay2021,
                MarineDay2022Onwards,
                MountainDayThrough2019,
                MountainDay2020,
                MountainDay2021,
                MountainDay2022Onwards,
                RespectForTheAgedDayThrough2002,
                RespectForTheAgedDay2003Onwards,
                HealthAndSportsDayThrough1999,
                HealthAndSportsDay2000OnwardsThrough2019,
                HealthAndSportsDay2020,
                HealthAndSportsDay2021,
                HealthAndSportsDay2022Onwards,
                CultureDay,
                LaborThanksgivingDay,
                EmperorAkihitoBirthday,
                EmperorNaruhitoBirthday,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                VernalEquinoxes,
                AutumnalEquinoxes,
                CitizensHolidaySilverWeek,
                Misc2019Holidays,
                EquityTradingSystemFailure,
            )
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas.tseries.holiday import MO, TH, DateOffset, EasterMonday, GoodFriday, Holiday
from pytz import timezone

from .common_holidays import (
    ascension_day,
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    maundy_thursday,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

# This falls on the first Thursday after 18 April.
FirstDayOfSummer = Holiday(
    "First Day of Summer",
    month=4,
    day=19,
    offset=DateOffset(weekday=TH(1)),
)

LabourDay = european_labour_day()

MaundyThursday = maundy_thursday()
AscensionDay = ascension_day()
WhitMonday = whit_monday()

NationalDay = Holiday("Icelandic Republic Day", month=6, day=17)

# This falls on the first Monday of August.
CommerceDay = Holiday(
    "Commerce Day",
    month=8,
    day=1,
    offset=DateOffset(weekday=MO(1)),
)

ChristmasEve = christmas_eve()
Christmas = christmas()
BoxingDay = boxing_day()

NewYearsEve = new_years_eve()


class XICEExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Iceland Exchange.

    Open Time: 9:30 AM, GMT (Greenwich Mean Time)
    Close Time: 3:30 PM, GMT (Greenwich Mean Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Maundy Thursday
      - Good Friday
      - Easter Monday
      - First Day of Summer
      - Labour Day
      - Ascension Day
      - Whit Monday
      - National Day
      - Commerce Day
      - Christmas Eve
      - Christmas Day
      - Boxing Day
      - New Year's Eve

    Early Closes:
      - None
    """

    name = "XICE"
    tz = timezone("Atlantic/Reykjavik")
    open_times = ((None, time(9, 30)),)
    close_times = ((None, time(15, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                MaundyThursday,
                GoodFriday,
                EasterMonday,
                FirstDayOfSummer,
                LabourDay,
                AscensionDay,
                WhitMonday,
                NationalDay,
                CommerceDay,
                ChristmasEve,
                Christmas,
                BoxingDay,
                NewYearsEve,
            ]
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas.tseries.holiday import (
    EasterMonday,
    GoodFriday,
    Holiday,
    previous_friday,
    previous_workday,
)
from pytz import timezone

from .common_holidays import (
    all_saints_day,
    ascension_day,
    assumption_day,
    christmas,
    christmas_eve,
    corpus_christi,
    epiphany,
    european_labour_day,
    immaculate_conception,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

Epiphany = epiphany(end_date="2019")

AscensionDay = ascension_day(end_date="2019")
WhitMonday = whit_monday()
CorpusChristi = corpus_christi(end_date="2019")

LabourDay = european_labour_day()

AssumptionDay = assumption_day(end_date="2019")

NationalHoliday = Holiday("National Holiday", month=10, day=26)

AllSaintsDay = all_saints_day(end_date="2019")

ImmaculateConception = immaculate_conception(end_date="2019")

ChristmasEve = christmas_eve()
Christmas = christmas()

SaintStephensDay = Holiday("Saint Stephen's Day", month=12, day=26)

# Prior to 2016, when New Year's Eve fell on the weekend, it was observed
# on the preceding Friday. In 2016 and after, it is not made up.
NewYearsEveThrough2015 = new_years_eve(
    observance=previous_friday,
    end_date="2016",
)
NewYearsEve2016Onwards = new_years_eve(start_date="2016")

# Early Closes
# ------------
# The last weekday before Dec 31 is an early close starting in 2017. Note: this
# is decided upon separately each year somewhen in November/December.
# Official announcements can be found at
#
#    https://www.wienerborse.at/en/news/vienna-stock-exchange-news-list/
#
# when searching for "Last trading day".
LastWorkingDay = Holiday(
    "Last Working Day of Year Early Close",
    month=12,
    day=31,
    start_date="2017-01-01",
    observance=previous_workday,
)


class XWBOExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Wiener Borse AG exchange in Vienna, Austria.

    Open Time: 9:00 AM, CET (Central European Time)
    Close Time: 5:30 PM, CET (Central European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Epiphany
      - Good Friday
      - Easter Monday
      - Ascension Day
      - Whit Monday
      - Corpus Christi
      - Labour Day
      - Assumption Day
      - National Holiday
      - All Saints Day
      - Immaculate Conception
      - Christmas Eve
      - Christmas Day
      - St. Stephen's Day
      - New Year's Eve

    Early Closes:
      - Last working day before Dec. 31
    """

    regular_early_close = time(14, 15)

    name = "XWBO"

    tz = timezone("Europe/Vienna")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                Epiphany,
                GoodFriday,
                EasterMonday,
                AscensionDay,
                WhitMonday,
                CorpusChristi,
                LabourDay,
                AssumptionDay,
                NationalHoliday,
                AllSaintsDay,
                ImmaculateConception,
                ChristmasEve,
                Christmas,
                SaintStephensDay,
                NewYearsEveThrough2015,
                NewYearsEve2016Onwards,
            ]
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        LastWorkingDay,
                    ]
                ),
            )
        ]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pytz import UTC, timezone

from .tase_holidays import (
    FastDay,
    IndependenceDay,
    MemorialDay,
    NewYear,
    NewYear2,
    NewYearsEve,
    Passover,
    Passover2,
    PassoverInterimDay1,
    PassoverInterimDay2,
    PassoverInterimDay3,
    PassoverInterimDay4,
    Passover2Eve,
    PassoverEve,
    Pentecost,
    PentecostEve,
    Purim,
    SimchatTorah,
    SimchatTorahEve,
    Sukkoth,
    SukkothEve,
    YomKippur,
    YomKippurEve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# All holidays are defined as ad-hoc holidays for each year since there is
# currently no support for Hebrew calendar holiday rules in pandas.


class XTAEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for TASE Stock Exchange.

    Open/close times are continuous trading times valid Mon-Thu for Shares
    Group A. Trading schedule differs on Sundays.

    Open Time: 9:59 AM, Asia/Tel_Aviv (randomly between 9:59 and 10:00).
    Close Time: 5:14 PM, Asia/Tel_Aviv (randomly between 5:14 and 5:15).

    Regularly-Observed Holidays (not necessarily in order):
    - Purim
    - Passover Eve
    - Passover
    - Passover II Eve
    - Passover II
    - Memorial Day
    - Independence Day
    - Pentecost Eve
    - Pentecost
    - Tisha B'Av
    - New Year's Eve
    - New Year
    - New Year II
    - Yom Kippur Eve
    - Yom Kippur
    - Sukkoth Eve
    - Sukkoth
    - Simchat Torah Eve
    - Simchat Torah

    Note these dates are only checked against 2019-2023.

    https://info.tase.co.il/eng/about_tase/corporate/pages/vacation_schedule.aspx

    Daylight Saving Time in Israel comes into effect on the Friday before the
    last Sunday in March, and lasts until the last Sunday in October. During the
    Daylight Saving time period the clock will be UTC+3, and UTC+2 for the rest
    of the year.
    """  # noqa

    start_date = pd.Timestamp("2019-01-01", tz=UTC)

    name = "XTAE"

    tz = timezone("Asia/Tel_Aviv")

    open_times = ((None, time(10, 0)),)

    close_times = ((None, time(17, 15)),)

    regular_early_close = time(14, 15)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                Purim,
                PassoverEve,
                Passover,
                Passover2Eve,
                Passover2,
                MemorialDay,
                IndependenceDay,
                PentecostEve,
                Pentecost,
                FastDay,
                NewYearsEve,
                NewYear,
                NewYear2,
                YomKippurEve,
                YomKippur,
                SukkothEve,
                Sukkoth,
                SimchatTorahEve,
                SimchatTorah,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            # 2019
            # Election Day
            pd.Timestamp("2019-04-09"),
            # Election Day
            pd.Timestamp("2019-09-17"),
            # 2020
            # Election Day
            pd.Timestamp("2020-03-02"),
            # 2021
            # Election Day
            pd.Timestamp("2021-03-23"),
        ]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        PassoverInterimDay1,
                        PassoverInterimDay2,
                        PassoverInterimDay3,
                        PassoverInterimDay4,
                    ]
                )
            )
        ]

    @property
    def weekmask(self):
        return "1111001"



# https://github.com/pandas-dev/pandas/blob/master/pandas/tseries/offsets.py
# https://github.com/pandas-dev/pandas/blob/master/pandas/_libs/tslibs/offsets.pyx

from datetime import datetime, timedelta

import toolz
import numpy as np
import pandas as pd

from pandas.tseries.offsets import CustomBusinessDay
from pandas._libs.tslibs.offsets import apply_wraps


class CompositeCustomBusinessDay(CustomBusinessDay):

    _prefix = "C"
    _attributes = tuple(
        [
            "n",
            "normalize",
            "weekmask",
            "holidays",
            "calendar",
            "offset",
            "business_days",
        ]
    )

    def __init__(
        self,
        n=1,
        normalize=False,
        weekmask="Mon Tue Wed Thu Fri",
        holidays=None,
        calendar=None,
        offset=timedelta(0),
        business_days=None,
    ):
        CustomBusinessDay.__init__(
            self, n, normalize, weekmask, holidays, calendar, offset
        )
        self.business_days = business_days

    def __setstate__(self, state):
        self.business_days = state.pop("business_days")
        CustomBusinessDay.__setstate__(self, state)

    @property
    def business_days(self):
        """
        Returns list of tuples of (start_date, end_date, custom_business_day)
        which overrides default behavior for the given interval, which starts
        from start_date to end_date, inclusive in both sides.
        """
        return tuple(self._business_days)

    @business_days.setter
    def business_days(self, business_days):
        self._business_days = []
        self._business_days_index = pd.IntervalIndex([], closed="both")
        self._business_days_all_index = pd.IntervalIndex([], closed="both")
        if business_days is not None:
            self._business_days = business_days
            business_days_intervals = []
            for start_date, end_date, _ in self._business_days:
                if start_date is None:
                    start_date = pd.Timestamp.min
                else:
                    start_date = pd.Timestamp(start_date)
                if end_date is None:
                    end_date = pd.Timestamp.max
                else:
                    end_date = pd.Timestamp(end_date)
                interval = pd.Interval(start_date, end_date, closed="both")
                business_days_intervals.append(interval)
            self._business_days_index = pd.IntervalIndex(
                business_days_intervals, closed="both"
            )
            business_days_all_intervals = []
            right = self._business_days_index[0]
            if pd.Timestamp.min < right.left:
                interval = pd.Interval(
                    pd.Timestamp.min, right.left - pd.Timedelta(1, unit="D")
                )
                business_days_all_intervals.append(interval)
            business_days_all_intervals.append(right)
            for left, right in toolz.sliding_window(
                2, toolz.concatv(self._business_days_index)
            ):
                if right.left - left.right > pd.Timestamp(1, unit="D"):
                    interval = pd.Interval(
                        left.right + pd.Timedelta(1, unit="D"),
                        right.left - pd.Timedelta(1, unit="D"),
                    )
                    business_days_all_intervals.append(interval)
                business_days_all_intervals.append(right)
            left = self._business_days_index[-1]
            if left.right < pd.Timestamp.max:
                interval = pd.Interval(
                    left.right + pd.Timedelta(1, unit="D"), pd.Timestamp.max
                )
                business_days_all_intervals.append(interval)
            self._business_days_all_index = pd.IntervalIndex(
                business_days_all_intervals, closed="both"
            )

    def _as_custom_business_day(self):
        return CustomBusinessDay(
            self.n,
            self.normalize,
            self.weekmask,
            self.holidays,
            self.calendar,
            self.offset,
        )

    def _custom_business_day_for(
        self, other, n=None, is_edge=False, with_interval=False
    ):
        loc = self._business_days_all_index.get_loc(other)
        if is_edge and n is not None:
            loc += np.sign(n)
        interval = self._business_days_all_index[loc]
        try:
            loc = self._business_days_index.get_loc(interval.left)
        except KeyError:
            bday = self._as_custom_business_day()
        else:
            bday = self._business_days[loc][-1]
        if n is not None:
            bday = bday.base * n
        if with_interval:
            return bday, interval
        else:
            return bday

    def _moved(self, from_date, to_date, bday):
        return np.busday_count(
            np.datetime64(from_date.date()),
            np.datetime64(to_date.date()),
            busdaycal=bday.calendar,
        )

    @apply_wraps
    def _apply(self, other):
        if isinstance(other, datetime):
            moved = 0
            remaining = self.n - moved
            bday, interval = self._custom_business_day_for(
                other, remaining, with_interval=True
            )
            result = bday.apply(other)
            while not interval.left <= result <= interval.right:
                previous_other = other
                if result < interval.left:
                    other = interval.left
                elif result > interval.right:
                    other = interval.right
                else:
                    raise RuntimeError("Should not reach here")
                moved += self._moved(previous_other, other, bday)
                remaining = self.n - moved
                if remaining == 0:
                    break
                bday, interval = self._custom_business_day_for(
                    other, remaining, is_edge=True, with_interval=True
                )
                result = bday.apply(other)
            return result
        else:
            return super().apply(other)

    # backwards compat
    apply = _apply

    def is_on_offset(self, dt):
        if self.normalize and not _is_normalized(dt):
            return False
        day64 = _to_dt64D(dt)
        bday = self._custom_business_day_for(day64)
        return np.is_busday(day64, busdaycal=bday.calendar)


def _is_normalized(dt):
    if dt.hour != 0 or dt.minute != 0 or dt.second != 0 or dt.microsecond != 0:
        # Regardless of whether dt is datetime vs Timestamp
        return False
    if isinstance(dt, pd.Timestamp):
        return dt.nanosecond == 0
    return True


def _to_dt64D(dt):
    # Currently
    # > np.datetime64(dt.datetime(2013,5,1),dtype='datetime64[D]')
    # numpy.datetime64('2013-05-01T02:00:00.000000+0200')
    # Thus astype is needed to cast datetime to datetime64[D]
    if getattr(dt, "tzinfo", None) is not None:
        # Get the nanosecond timestamp,
        # equiv `Timestamp(dt).value` or `dt.timestamp() * 10**9`
        dt = pd.Timestamp(dt).value
        dt = np.int64(dt).astype("datetime64[ns]")
    else:
        dt = np.datetime64(dt)
    if dt.dtype.name != "datetime64[D]":
        dt = dt.astype("datetime64[D]")
    return dt


def _get_calendar(weekmask, holidays, calendar):
    """
    Generate busdaycalendar
    """
    if isinstance(calendar, np.busdaycalendar):
        if not holidays:
            holidays = tuple(calendar.holidays)
        elif not isinstance(holidays, tuple):
            holidays = tuple(holidays)
        else:
            # Trust that calendar.holidays and holidays are
            # consistent
            pass
        # Update weekmask if applicable (added)
        calendar = np.busdaycalendar(weekmask, holidays)
        return calendar, holidays

    if holidays is None:
        holidays = []
    # Handle non list holidays also (added)
    if isinstance(holidays, pd.DatetimeIndex):
        holidays = holidays.tolist()
    try:
        holidays = holidays + calendar.holidays().tolist()
    except AttributeError:
        pass
    holidays = [_to_dt64D(dt) for dt in holidays]
    holidays = tuple(sorted(holidays))

    kwargs = {"weekmask": weekmask}
    if holidays:
        kwargs["holidays"] = holidays

    busdaycalendar = np.busdaycalendar(**kwargs)
    return busdaycalendar, holidays


class MultipleWeekmaskCustomBusinessDay(CompositeCustomBusinessDay):

    _prefix = "C"
    _attributes = tuple(
        [
            "n",
            "normalize",
            "weekmask",
            "holidays",
            "calendar",
            "offset",
            "business_days",
            "weekmasks",
        ]
    )

    def __init__(
        self,
        n=1,
        normalize=False,
        weekmask="Mon Tue Wed Thu Fri",
        holidays=None,
        calendar=None,
        offset=timedelta(0),
        business_days=None,
        weekmasks=None,
    ):
        self._weekmasks = weekmasks
        if business_days is None and weekmasks is not None:
            calendars = [
                _get_calendar(weekmask=weekmask, holidays=holidays, calendar=calendar)
                for _start_date, _end_date, weekmask in weekmasks
            ]
            business_days = [
                (
                    start_date,
                    end_date,
                    CustomBusinessDay(
                        n, normalize, weekmask, holidays, calendar, offset
                    ),
                )
                for (start_date, end_date, weekmask), (calendar, holidays) in zip(
                    weekmasks, calendars
                )
            ]
        CompositeCustomBusinessDay.__init__(
            self, n, normalize, weekmask, holidays, calendar, offset, business_days
        )

    def __setstate__(self, state):
        self.weekmasks = state.pop("weekmasks")
        CompositeCustomBusinessDay.__setstate__(self, state)

    @property
    def weekmasks(self):
        return tuple(self._weekmasks)

    @weekmasks.setter
    def weekmasks(self, weekmasks):
        self._weekmasks = weekmasks



from datetime import time
from itertools import chain

from pandas.tseries.holiday import (
    GoodFriday,
    USLaborDay,
    USPresidentsDay,
    USThanksgivingDay,
)
from pytz import timezone

from .exchange_calendar import ExchangeCalendar
from exchange_calendars.exchange_calendar import HolidayCalendar
from exchange_calendars.us_holidays import (
    Christmas,
    HurricaneSandyClosings,
    USBlackFridayInOrAfter1993,
    USIndependenceDay,
    USMartinLutherKingJrAfter1998,
    USMemorialDay,
    USNationalDaysofMourning,
    USNewYearsDay,
    USJuneteenth
)


class XCBFExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the CBOE Futures Exchange (XCBF).

    http://cfe.cboe.com/aboutcfe/expirationcalendar.aspx

    Open Time: 8:30am, America/Chicago
    Close Time: 3:15pm, America/Chicago

    (We are ignoring extended trading hours for now)
    """

    name = "XCBF"

    tz = timezone("America/Chicago")

    open_times = ((None, time(8, 30)),)

    close_times = ((None, time(15, 15)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                USNewYearsDay,
                USMartinLutherKingJrAfter1998,
                USPresidentsDay,
                GoodFriday,
                USIndependenceDay,
                USMemorialDay,
                USJuneteenth,
                USLaborDay,
                USThanksgivingDay,
                Christmas,
            ]
        )

    @property
    def special_closes(self):
        return [
            (
                time(12, 15),
                HolidayCalendar(
                    [
                        USBlackFridayInOrAfter1993,
                    ]
                ),
            )
        ]

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                HurricaneSandyClosings,
                USNationalDaysofMourning,
            )
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta

from pandas.tseries.holiday import Easter, EasterMonday, Holiday
from pandas.tseries.offsets import Day
from pytz import timezone

from .common_holidays import (
    all_saints_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import THURSDAY, TUESDAY, HolidayCalendar, ExchangeCalendar


def four_day_weekend(dt, include_mon: bool = True, include_fri: bool = True):
    """
    Custom observance function as for almost all holidays in the XBUD calendar,
    if the holiday falls on a Tuesday the previous Monday also becomes a holiday,
    and if the holiday falls on a Thursday the following Friday also becomes a holiday.

    Parameters
    ----------
    dt : datetime
         Unadjusted raw holiday datetimes
    include_mon : boolean
         If holiday falls on a Tuesday, previous Monday becomes a holiday.
    include_fri : boolean
         If holiday falls on a Thursday, following Friday becomes a holiday.
    """
    extras = []
    if include_mon:
        extras.append(dt[dt.weekday == TUESDAY] - timedelta(1))  # mv Tues back one day
    if include_fri:
        # mv Thurs ahead one day
        extras.append(dt[dt.weekday == THURSDAY] + timedelta(1))
    return dt.append(extras)


NewYearsDay = new_years_day(observance=four_day_weekend)

NationalHoliday1 = Holiday("National Day", month=3, day=15, observance=four_day_weekend)

# Need custom start year so can't use pandas GoodFriday
GoodFriday = Holiday(
    "Good Friday", month=1, day=1, offset=[Easter(), Day(-2)], start_date="2012"
)

LabourDay = european_labour_day(observance=four_day_weekend)

WhitMonday = whit_monday()

StStephensDay = Holiday(
    "St. Stephen's Day",
    month=8,
    day=20,
    observance=four_day_weekend,
)

NationalHoliday2 = Holiday(
    "National Day",
    month=10,
    day=23,
    observance=four_day_weekend,
)

AllSaintsDay = all_saints_day(observance=four_day_weekend)

# Christmas Eve does not follow the four day weekend rule
ChristmasEve = christmas_eve()

ChristmasDay = christmas()

# XBUD always has a holiday for the second day of Christmas (26th),
# but starting in 2013 if the 26th falls on a Thursday then the
# 27th (Friday) is also taken off
SecondDayOfChristmas = Holiday(
    "Second Day of Christmas (w/ no added Friday off)",
    month=12,
    day=26,
    end_date="2013",
)

SecondDayOfChristmasAddFriday = Holiday(
    "Second Day of Christmas (w/ added Friday off)",
    month=12,
    day=26,
    start_date="2013",
    # Don't apply the rule here where the previous Monday also becomes a holiday if this
    # falls onto a Tuesday.
    # Rationale: The previous Monday in this case is Christmas which is already defined
    # as a holiday above. Apply the other rule where the following Friday becomes a
    # holiday if it falls onto a Thursday as usual.
    observance=lambda dt: four_day_weekend(dt, include_mon=False, include_fri=True),
)

# Starting in 2011, New Year's Eve is observed as a holiday every year.
# In some cases pre-2011, the 31st becomes a holiday due to the four day
# weekend rule (when Jan 1 falls on a Tuesday).
# Also, when NYE starts being observed as a holiday it does NOT follow
# the four day weekend rule (no 30ths are holidays)
NewYearsEve = new_years_eve(start_date="2011")


class XBUDExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Budapest Stock Exchange (XBUD).

    Open Time: 9:00 AM, CET
    Close Time: 5:00 PM, CET

    Regularly-Observed Holidays:
    - New Year's Day
    - National Holiday (Mar 15)
    - Good Friday
    - Easter Monday
    - Labour Day (May 1)
    - Whit Monday (50 days after Easter Sunday)
    - St. Stephen's Day (Aug 20)
    - National Holiday (Oct 23)
    - All Saint's Day (Nov 1)
    - Christmas Eve
    - Christmas Day
    - Second Day of Christmas (Dec 26)
    - New Year's Eve

    Early Closes:
    - None
    """

    name = "XBUD"

    tz = timezone("Europe/Budapest")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 00)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NationalHoliday1,
                GoodFriday,
                EasterMonday,
                LabourDay,
                WhitMonday,
                StStephensDay,
                NationalHoliday2,
                AllSaintsDay,
                ChristmasEve,
                ChristmasDay,
                SecondDayOfChristmas,
                SecondDayOfChristmasAddFriday,
                NewYearsEve,
            ]
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .calendar_utils import (
    clear_calendars,
    deregister_calendar,
    get_calendar,
    get_calendar_names,
    register_calendar,
    register_calendar_alias,
    register_calendar_type,
    resolve_alias,
    names_to_aliases,
    aliases_to_names,
)
from .exchange_calendar import ExchangeCalendar

__all__ = [
    "clear_calendars",
    "deregister_calendar",
    "get_calendar",
    "get_calendar_names",
    "register_calendar",
    "register_calendar_alias",
    "register_calendar_type",
    "resolve_alias",
    "names_to_aliases",
    "aliases_to_names",
    "ExchangeCalendar",
]

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

import pandas as pd
from pandas.tseries.holiday import (
    MO,
    DateOffset,
    EasterMonday,
    GoodFriday,
    Holiday,
    previous_friday,
    weekend_to_monday,
)
import pytz

from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    weekend_boxing_day,
    weekend_christmas,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day(observance=weekend_to_monday)

StPatricksDay = Holiday(
    "St. Patrick's Day",
    month=3,
    day=17,
    end_date="2001",
    observance=weekend_to_monday,
)

LabourDayTo2009 = european_labour_day(
    start_date="2008",
    end_date="2010",
    observance=weekend_to_monday,
)

LabourDayFrom2019 = european_labour_day(
    start_date="2019",
    observance=weekend_to_monday,
)

MayBankHoliday = Holiday(
    "May Bank Holiday",
    month=5,
    day=1,
    end_date="2019",
    offset=DateOffset(weekday=MO(1)),
)
JuneBankHoliday = Holiday(
    "June Bank Holiday",
    month=6,
    day=1,
    end_date="2019",
    offset=DateOffset(weekday=MO(1)),
)

LastTradingDayBeforeChristmas = Holiday(
    "Last Trading Day Before Christmas",
    month=12,
    day=24,
    start_date="2010",
    observance=previous_friday,
)

ChristmasEve = christmas_eve(end_date="2005")
Christmas = christmas()
WeekendChristmas = weekend_christmas()
BoxingDay = boxing_day()
WeekendBoxingDay = weekend_boxing_day()

LastTradingDayOfCalendarYear = Holiday(
    "Last Trading Day Of Calendar Year",
    month=12,
    day=31,
    start_date="2010",
    observance=previous_friday,
)

# Ad hoc closes.
March1BadWeather = pd.Timestamp("2018-03-01")
# Ad hoc holidays.
March2BadWeather = pd.Timestamp("2018-03-02")


class XDUBExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Irish Stock Exchange in Dublin.

    Open Time: 8:00 AM, GMT
    Close Time: 4:28 PM, GMT

    Regularly-Observed Holidays:
      - New Years Day
      - Good Friday
      - Easter Monday
      - May Bank Holiday
      - June Bank Holiday
      - Christmas Day
      - St. Stephen's Day (Boxing Day)

    Holidays No Longer Observed:
      - St. Patrick's Day
      - Labour Day

    Early Closes:
      - Christmas Eve
      - New Year's Eve
    """

    name = "XDUB"
    tz = pytz.timezone("Europe/Dublin")
    open_times = ((None, time(8)),)
    close_times = ((None, time(16, 28)),)
    regular_early_close = time(12, 28)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                StPatricksDay,
                GoodFriday,
                EasterMonday,
                LabourDayTo2009,
                LabourDayFrom2019,
                MayBankHoliday,
                JuneBankHoliday,
                ChristmasEve,
                Christmas,
                WeekendChristmas,
                BoxingDay,
                WeekendBoxingDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [March2BadWeather]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        LastTradingDayBeforeChristmas,
                        LastTradingDayOfCalendarYear,
                    ]
                ),
            ),
        ]

    @property
    def special_closes_adhoc(self):
        return [(self.regular_early_close, pd.DatetimeIndex([March1BadWeather]))]



from datetime import time

import numpy as np
import pandas as pd
from pandas.tseries.holiday import GoodFriday
from pytz import UTC, timezone

from exchange_calendars.exchange_calendar import ExchangeCalendar, HolidayCalendar
from exchange_calendars.us_holidays import Christmas, USNewYearsDay

# Number of hours of offset between the open and close times dictated by this
# calendar versus the 6:31am to 5:00pm times over which we want to simulate
# futures algos.
FUTURES_OPEN_TIME_OFFSET = 12.5
FUTURES_CLOSE_TIME_OFFSET = -1


class QuantopianUSFuturesCalendar(ExchangeCalendar):
    """Synthetic calendar for trading US futures.

    This calendar is a superset of all of the US futures exchange
    calendars provided by Zipline (CFE, CME, ICE), and is intended for
    trading across all of these exchanges.

    Notes
    -----
    Open Time: 6:00 PM, America/New_York
    Close Time: 6:00 PM, America/New_York

    Regularly-Observed Holidays:
    - New Years Day
    - Good Friday
    - Christmas

    In order to align the hours of each session, we ignore the Sunday
    CME Pre-Open hour (5-6pm).
    """

    name = "us_futures"
    tz = timezone("America/New_York")
    open_times = ((None, time(18)),)
    close_times = ((None, time(18)),)
    open_offset = -1

    def execution_time_from_open(self, open_dates):
        return open_dates + pd.Timedelta(hours=FUTURES_OPEN_TIME_OFFSET)

    def execution_time_from_close(self, close_dates):
        return close_dates + pd.Timedelta(hours=FUTURES_CLOSE_TIME_OFFSET)

    def execution_minutes_for_session(
        self, session_label: pd.DatetimeIndex
    ) -> pd.DatetimeIndex:
        """
        Given a session label, return the execution minutes for that session.

        Parameters
        ----------
        session_label
            A session label whose session's minutes are desired.

        Returns
        -------
        pd.DateTimeIndex
            All the execution minutes for the given session.
        """
        start = self.execution_time_from_open(self.session_first_minute(session_label))
        end = self.execution_time_from_close(self.session_last_minute(session_label))
        return self.minutes_in_range(start=start, end=end)

    def execution_minutes_for_sessions_in_range(self, start, stop):
        minutes = self.execution_minutes_for_session
        return pd.DatetimeIndex(
            np.concatenate(
                [minutes(session) for session in self.sessions_in_range(start, stop)]
            ),
            tz=UTC,
        )

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                USNewYearsDay,
                GoodFriday,
                Christmas,
            ]
        )



from datetime import date, datetime

import pandas as pd
from pandas._libs.tslibs.conversion import localize_pydatetime
from pandas.tseries.holiday import Day, Holiday
from pandas.tseries.offsets import Easter
from pyluach import dates, hebrewcal

try:
    from pandas._libs.tslibs.offsets import apply_wraps
except ImportError:
    from pandas.tseries.offsets import apply_wraps

# Auxiliary functions to get Hebrew dates for holidays observed by TASE for a
# given Hebrew calendar year. These are just the raw dates with no adjustments
# applied.
#
# Note: pyluach uses the biblical month numbering scheme where the year is
# incremented when moving from the month of Elul (6) to Tishrei (7). See also
# https://en.wikipedia.org/wiki/Hebrew_calendar.


def _purim(year):
    """
    Return the Hebrew date for Purim in the given Hebrew year.
    """
    # Purim is observed in Adar (12), or Adar II (13) if in a leap year.
    return dates.HebrewDate(year.year, 13 if year.leap else 12, 14)


def _passover(year):
    """
    Return the Hebrew date for the first day of Passover in the given Hebrew
    year.
    """
    return dates.HebrewDate(year.year, 1, 15)


def _memorial_day(year):
    """
    Return the Hebrew date for Memorial Day in the given Hebrew year.

    Note: Independence Day is always celebrated the following day.
    """
    return dates.HebrewDate(year.year, 2, 4)


def _pentecost(year):
    """
    Return the Hebrew date for Pentecost in the given Hebrew year.
    """
    return dates.HebrewDate(year.year, 3, 6)


def _fast_day(year):
    """
    Return the Hebrew date for Tisha B'Av in the given Hebrew year.
    """
    return dates.HebrewDate(year.year, 5, 9)


def _new_year(year):
    """
    Return the Hebrew date for the first day of a new year in the given Hebrew
    year.
    """
    return dates.HebrewDate(year.year, 7, 1)


def _yom_kippur(year):
    """
    Return the Hebrew date for Yom Kippur in the given Hebrew year.
    """
    return dates.HebrewDate(year.year, 7, 10)


def _sukkoth(year):
    """
    Return the Hebrew date for Sukkoth in the given Hebrew year.
    """
    return dates.HebrewDate(year.year, 7, 15)


def _simchat_torah(year):
    """
    Return the Hebrew date for Simchat Torah in the given Hebrew year.
    """
    return dates.HebrewDate(year.year, 7, 22)


def _hebrew_year(year):
    """
    Return the Hebrew calendar year that corresponds to 1st January of the
    given Gregorian calendar year.

    1st January of any Gregorian calendar year, say x, always falls into the
    month of Tevet (10) or Shevat (11) of some Hebrew year f(x). Also, we have
    f(x+1) = f(x) + 1, so that any year in the Gregorian calendar always
    overlaps with two consecutive years in the Hebrew calendar and vice versa.
    """
    return hebrewcal.Year(dates.GregorianDate(year, 1, 1).to_heb().year)


# Auxilliary functions to calculate Gregorian dates for holidays observed by
# TASE for a given Gregorian calendar year. Adjustments are also applied.


def purim(year):
    """
    Return the Gregorian date for Purim in the given Gregorian calendar year.
    """
    return _purim(_hebrew_year(year)).to_greg()


def passover(year):
    """
    Return the Gregorian date for the first day of Passover in the given
    Gregorian calendar year.
    """
    return _passover(_hebrew_year(year)).to_greg()


def memorial_day(year):
    """
    Return the Gregorian date for Memorial Day in the given Gregorian calendar
    year.
    """

    # Regular Memorial Day date.
    d = _memorial_day(_hebrew_year(year)).to_greg()

    # Reschedule to avoid Sabbath desecration, maybe.
    if d.weekday() == 5:
        # Falls on a Thursday, so Independency Day falls on the Friday.
        # Moved down by one day.
        return d - 1
    elif d.weekday() == 6:
        # Falls on a Friday, so Independence Day falls on the Saturday.
        # Moved down by two days.
        return d - 2
    elif d.weekday() == 7:
        # Falls on a Saturday, therefore moved up by one day.
        return d + 1
    else:
        return d


def pentecost(year):
    """
    Return the Gregorian date for Pentecost in the given Gregorian calendar
    year.
    """
    return _pentecost(_hebrew_year(year)).to_greg()


def fast_day(year):
    """
    Return the Gregorian date for Tisha B'Av in the given Gregorian calendar
    year.
    """
    d = _fast_day(_hebrew_year(year)).to_greg()

    # Reschedule if it falls on Sabbath (Saturday), maybe.
    if d.weekday() == 7:
        # Falls on a Saturday, therefore moved up by one day.
        return d + 1
    else:
        return d


def new_year(year):
    """
    Return the Gregorian date for the first day of a new year in the given
    Gregorian calendar year.
    """
    return _new_year(_hebrew_year(year + 1)).to_greg()


def yom_kippur(year):
    """
    Return the Gregorian date for Yom Kippur in the given Gregorian calendar
    year.
    """
    return _yom_kippur(_hebrew_year(year + 1)).to_greg()


def sukkoth(year):
    """
    Return the Gregorian date for Sukkoth in the given Gregorian calendar year.
    """
    return _sukkoth(_hebrew_year(year + 1)).to_greg()


def simchat_torah(year):
    """
    Return the Gregorian date for Simchat Torah in the given Gregorian calendar
    year.
    """
    return _simchat_torah(_hebrew_year(year + 1)).to_greg()


def _is_normalized(dt):
    if dt.hour != 0 or dt.minute != 0 or dt.second != 0 or dt.microsecond != 0:
        # Regardless of whether dt is datetime vs Timestamp
        return False
    if isinstance(dt, pd.Timestamp):
        return dt.nanosecond == 0
    return True


class _HolidayOffset(Easter):
    """
    Auxiliary class for DateOffset instances for the different holidays.
    """

    @property
    def holiday(self):
        """
        Return the Gregorian date for the holiday in a given Gregorian calendar
        year.
        """
        pass

    @apply_wraps
    def _apply(self, other):
        current = self.holiday(other.year).to_pydate()
        current = datetime(current.year, current.month, current.day)
        current = localize_pydatetime(current, other.tzinfo)

        n = self.n
        if n >= 0 and other < current:
            n -= 1
        elif n < 0 and other > current:
            n += 1
        # TODO: Why does this handle the 0 case the opposite of others?

        # NOTE: self.holiday a dates.GregorianDate so we have to convert to
        # type of other
        new = self.holiday(other.year + n).to_pydate()
        new = datetime(
            new.year,
            new.month,
            new.day,
            other.hour,
            other.minute,
            other.second,
            other.microsecond,
        )
        return new

    # backwards compat
    apply = _apply

    def is_on_offset(self, dt):
        if self.normalize and not _is_normalized(dt):
            return False
        return date(dt.year, dt.month, dt.day) == self.holiday(dt.year).to_pydate()


# DateOffset subclasses for holidays observed by TASE.


class _Purim(_HolidayOffset):
    @property
    def holiday(self):
        return purim


class _Passover(_HolidayOffset):
    @property
    def holiday(self):
        return passover


class _MemorialDay(_HolidayOffset):
    @property
    def holiday(self):
        return memorial_day


class _Pentecost(_HolidayOffset):
    @property
    def holiday(self):
        return pentecost


class _FastDay(_HolidayOffset):
    @property
    def holiday(self):
        return fast_day


class _NewYear(_HolidayOffset):
    @property
    def holiday(self):
        return new_year


class _YomKippur(_HolidayOffset):
    @property
    def holiday(self):
        return yom_kippur


class _Sukkoth(_HolidayOffset):
    @property
    def holiday(self):
        return sukkoth


class _SimchatTorah(_HolidayOffset):
    @property
    def holiday(self):
        return simchat_torah


# Holiday instances for holidays observed by TASE.
Purim = Holiday("Purim", month=1, day=1, offset=[_Purim()])
PassoverEve = Holiday("Passover Eve", month=1, day=1, offset=[_Passover(), Day(-1)])
Passover = Holiday("Passover", month=1, day=1, offset=[_Passover()])
Passover2Eve = Holiday("Passover II Eve", month=1, day=1, offset=[_Passover(), Day(5)])
Passover2 = Holiday("Passover II", month=1, day=1, offset=[_Passover(), Day(6)])
PentecostEve = Holiday("Pentecost Eve", month=1, day=1, offset=[_Pentecost(), Day(-1)])
Pentecost = Holiday("Pentecost", month=1, day=1, offset=[_Pentecost()])
FastDay = Holiday("Tisha B'Av", month=1, day=1, offset=[_FastDay()])
MemorialDay = Holiday("Memorial Day", month=1, day=1, offset=[_MemorialDay()])
IndependenceDay = Holiday(
    "Independence Day", month=1, day=1, offset=[_MemorialDay(), Day(1)]
)
NewYearsEve = Holiday("New Year's Eve", month=1, day=1, offset=[_NewYear(), Day(-1)])
NewYear = Holiday("New Year", month=1, day=1, offset=[_NewYear()])
NewYear2 = Holiday("New Year II", month=1, day=1, offset=[_NewYear(), Day(1)])
YomKippurEve = Holiday("Yom Kippur Eve", month=1, day=1, offset=[_YomKippur(), Day(-1)])
YomKippur = Holiday("Yom Kippur", month=1, day=1, offset=[_YomKippur()])
SukkothEve = Holiday("Sukkoth Eve", month=1, day=1, offset=[_Sukkoth(), Day(-1)])
Sukkoth = Holiday("Sukkoth", month=1, day=1, offset=[_Sukkoth()])
SimchatTorahEve = Holiday(
    "Simchat Torah Eve", month=1, day=1, offset=[_SimchatTorah(), Day(-1)]
)
SimchatTorah = Holiday("Simchat Torah", month=1, day=1, offset=[_SimchatTorah()])

# Passover interim days are the days between beginning and end of passover. Any otherwise regular business day in that
# period becomes an early close day.
PassoverInterimDay1 = Holiday(
    "Passover Interim Day",
    month=1,
    day=1,
    offset=[_Passover(), Day(1)],
    days_of_week=(0, 1, 2, 3, 6),
)
PassoverInterimDay2 = Holiday(
    "Passover Interim Day",
    month=1,
    day=1,
    offset=[_Passover(), Day(2)],
    days_of_week=(0, 1, 2, 3, 6),
)
PassoverInterimDay3 = Holiday(
    "Passover Interim Day",
    month=1,
    day=1,
    offset=[_Passover(), Day(3)],
    days_of_week=(0, 1, 2, 3, 6),
)
PassoverInterimDay4 = Holiday(
    "Passover Interim Day",
    month=1,
    day=1,
    offset=[_Passover(), Day(4)],
    days_of_week=(0, 1, 2, 3, 6),
)



from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    next_monday,
    nearest_workday,
    next_workday,
)
import pytz

from .common_holidays import new_years_day, eid_al_adha_first_day
from .exchange_calendar import (
    HolidayCalendar,
    ExchangeCalendar,
)


NewYearsDay = new_years_day()

NewYearHoliday = Holiday(
    "New Year Holiday",
    month=1,
    day=2,
)

OrthodoxChristmasDay = Holiday(
    "Orthodox Christmas Day",
    month=1,
    day=7,
)

InternationalWomensDay = Holiday(
    "International Women's Day",
    month=3,
    day=8,
    observance=nearest_workday,
)

NauryzHoliday1 = Holiday(
    "Nauryz Holiday",
    month=3,
    day=21,
    observance=next_monday,
)

NauryzHoliday2 = Holiday(
    "Nauryz Holiday",
    month=3,
    day=21,
    observance=lambda dt: next_workday(next_monday(dt)),
)

NauryzHoliday3 = Holiday(
    "Nauryz Holiday",
    month=3,
    day=21,
    observance=lambda dt: next_workday(next_workday(next_monday(dt))),
)

KazakhstanPeopleSolidarityDay = Holiday(
    "Kazakhstan People Solidarity Day",
    month=5,
    day=1,
    observance=next_monday,
)

DefendersDay = Holiday(
    "Defender's Day",
    month=5,
    day=7,
    observance=next_monday,
    start_date=pd.Timestamp("2013-01-01"),
)

VictoryDayHoliday = Holiday(
    "Victory Day Holiday",
    month=5,
    day=9,
    observance=nearest_workday,
)

CapitalCityDay = Holiday(
    "Capital City Day",
    month=7,
    day=6,
    observance=next_monday,
)

ConstitutionDay = Holiday(
    "Constitution Day",
    month=8,
    day=30,
    observance=next_monday,
)

FirstPresidentDay = Holiday(
    "First President Day",
    month=12,
    day=1,
    observance=next_monday,
    start_date=pd.Timestamp("2013-01-01"),
)

IndependenceDay = Holiday(
    "Independence Day",
    month=12,
    day=16,
    observance=next_monday,
)

IndependenceDayHoliday = Holiday(
    "Independence Day",
    month=12,
    day=17,
    observance=next_monday,
)


class AIXKExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Astana International Exchange (XIST).
                Available here: https://www.aix.kz/trading/trading-calendar/


    Regularly-Observed Holidays:
      - New Year's Day
      - New Year Holiday
      - Orthodox Christmas Day
      - International Women's Day
      - Nauryz Holiday
      - Nauryz Holiday
      - Nauryz Holiday
      - Kazakhstan People Solidarity Day
      - Defender’s Day
      - Victory Day Holiday
      - Capital City Day
      - Capital City Day
      - Kurban Ait Holiday (Eid-al-Adha)
      - Constitution Day
      - First President Day
      - Independence Day
      - Independence Day Holiday
    Early Closes:
      - None
    """

    name = "AIXK"

    tz = pytz.timezone("Asia/Almaty")

    open_times = ((None, time(11)),)

    close_times = ((None, time(17, 00)),)

    @classmethod
    def bound_start(cls) -> pd.Timestamp:
        return pd.Timestamp("2017-01-01")

    def _bound_start_error_msg(self, start: pd.Timestamp) -> str:
        msg = super()._bound_start_error_msg(start)
        return msg + f" (The exchange {self.name} was founded in 2017.)"

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NewYearHoliday,
                OrthodoxChristmasDay,
                InternationalWomensDay,
                NauryzHoliday1,
                NauryzHoliday2,
                NauryzHoliday3,
                KazakhstanPeopleSolidarityDay,
                DefendersDay,
                VictoryDayHoliday,
                CapitalCityDay,
                ConstitutionDay,
                FirstPresidentDay,
                IndependenceDay,
                IndependenceDayHoliday,
            ]
        )

    @property
    def adhoc_holidays(self):
        # It is common in Kazakhstan to have holidays also on days
        # between regular holiday and weekend
        misc_holidays = [
            # Bridge Day between Women's day - Weekend
            pd.Timestamp("2018-03-09"),
            # Bridge Day between Weekend - Kazakhstan People Solidarity Day
            pd.Timestamp("2018-04-30"),
            # Bridge Day between Defender's Day - Victory Day
            pd.Timestamp("2018-05-08"),
            # Bridge Day between Constitution Day - Weekend
            pd.Timestamp("2018-08-31"),
            # Bridge Day between New Year's Eve - New Year's day
            pd.Timestamp("2018-12-31"),
            # Bridge Day between Victory Day - Weekend
            pd.Timestamp("2019-05-10"),
            # Bridge Day between New Year's day - Weekend
            pd.Timestamp("2020-01-03"),
            # Bridge Day between Independence day - Weekend
            pd.Timestamp("2020-12-18"),
            # Bridge Day between Weekend - Capital City day
            pd.Timestamp("2021-06-05"),
            # New Year holiday
            pd.Timestamp("2022-01-03"),
            pd.Timestamp("2022-01-04"),
            # Bridge Day between Weekend - Women's day
            pd.Timestamp("2022-03-07"),
            # Defender's day
            pd.Timestamp("2022-05-10"),
            # Bridge Day between Weekend - Constitution day
            pd.Timestamp("2022-08-29"),
        ]
        return list(chain(misc_holidays, eid_al_adha_first_day))



from datetime import time

from pytz import UTC

from .exchange_calendar import ExchangeCalendar


class WeekdayCalendar(ExchangeCalendar):
    """
    A ExchangeCalendar for an exchange that is open every minute of every
    weekday.
    """

    name = "24/5"
    tz = UTC
    open_times = ((None, time(0)),)
    close_times = ((None, time(0)),)
    close_offset = 1



from datetime import time

from pytz import UTC

from .exchange_calendar import ExchangeCalendar


class AlwaysOpenCalendar(ExchangeCalendar):
    """A ExchangeCalendar for an exchange that's open every minute of every day."""

    name = "24/7"
    tz = UTC
    weekmask = "1111111"
    open_times = ((None, time(0)),)
    close_times = ((None, time(0, 0)),)
    close_offset = 1



#
# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import typing
import pandas as pd

if typing.TYPE_CHECKING:
    from exchange_calendars import ExchangeCalendar


class CalendarError(Exception):
    msg = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __str__(self):
        msg = self.msg.format(**self.kwargs)
        return msg

    __unicode__ = __str__
    __repr__ = __str__


class InvalidCalendarName(CalendarError):
    """
    Raised when a calendar with an invalid name is requested.
    """

    msg = "The requested ExchangeCalendar, {calendar_name}, does not exist."


class CalendarNameCollision(CalendarError):
    """
    Raised when the static calendar registry already has a calendar with a
    given name.
    """

    msg = "A calendar with the name {calendar_name} is already registered."


class CyclicCalendarAlias(CalendarError):
    """
    Raised when calendar aliases form a cycle.
    """

    msg = "Cycle in calendar aliases: [{cycle}]"


class NoSessionsError(CalendarError):
    """Raised if a requested calendar would have no sessions.

    NoSessionsError should be raised if `get_calendar` `start` and `end`
    parameters are passed as dates between and inclusive of which there are
    no sessions for the given calendar.
    """

    msg = (
        "The requested ExchangeCalendar, {calendar_name}, cannot be created as"
        " there would be no sessions between the requested `start` ('{start}')"
        " and `end` ('{end}') dates."
    )


class NotSessionError(ValueError):
    """Input does not represent a valid session.

    Raised if parameter expecting a session label receives input that
    parses correctly (UTC midnight) although is not a session.

    Parameters
    ----------
    calendar
        Calendar for which `ts` assumed as a session.

    ts
        Timestamp assumed as a session.

    param_name
        Name of a parameter that was to receive a session label.
    """

    def __init__(self, calendar: ExchangeCalendar, ts: pd.Timestamp, param_name: str):
        self.calendar = calendar
        self.ts = ts
        self.param_name = param_name

    def __str__(self) -> str:
        msg = (
            f"Parameter `{self.param_name}` takes a session"
            f" although received input that parsed to '{self.ts}' which"
        )

        if self.ts < self.calendar.first_session:
            msg += (
                " is earlier than the first session of calendar"
                f" '{self.calendar.name}' ('{self.calendar.first_session}')."
            )
        elif self.ts > self.calendar.last_session:
            msg += (
                " is later than the last session of calendar"
                f" '{self.calendar.name}' ('{self.calendar.last_session}')."
            )
        else:
            msg += f" is not a session of calendar '{self.calendar.name}'."
        return msg


class DateOutOfBounds(ValueError):
    """A date required to be within sessions' bounds is not.

    Parameters
    ----------
    calendar
        Calendar for which `date` required to be within sessions'
        bounds.

    date
        Date required to be within `calendar`'s sessions' bounds.

    param_name
        Name of a parameter receiving date.
    """

    def __init__(self, calendar: ExchangeCalendar, date: pd.Timestamp, param_name: str):
        self.calendar = calendar
        self.date = date
        self.param_name = param_name

    def __str__(self) -> str:
        msg = f"Parameter `{self.param_name}` receieved as '{self.date}' although"
        if self.date < self.calendar.first_session:
            msg += (
                " cannot be earlier than the first session of calendar"
                f" '{self.calendar.name}' ('{self.calendar.first_session}')."
            )
        elif self.date > self.calendar.last_session:
            msg += (
                " cannot be later than the last session of calendar"
                f" '{self.calendar.name}' ('{self.calendar.last_session}')."
            )
        else:
            assert (
                self.date < self.calendar.first_session
                or self.date > self.calendar.last_session
            )
        return msg


class NotTradingMinuteError(ValueError):
    """A timestamp assumed as a trading minute is not a trading minute.

    Parameters
    ----------
    calendar
        Calendar for which `minute` assumed as a trading minute.

    minute
        Minute assumed as a trading minute.

    param_name
        Name of a parameter that was to receive a trading minute.
    """

    def __init__(
        self,
        calendar: ExchangeCalendar,
        minute: pd.Timestamp,
        param_name: str,
    ):
        self.calendar = calendar
        self.minute = minute
        self.param_name = param_name

    def __str__(self) -> str:
        msg = (
            f"Parameter `{self.param_name}` takes a trading minute although"
            f" received input that parsed to '{self.minute}' which"
        )
        if self.minute < self.calendar.first_minute:
            msg += (
                " is earlier than the first trading minute of calendar"
                f" '{self.calendar.name}' ('{self.calendar.first_session}')."
            )
        elif self.minute > self.calendar.last_minute:
            msg += (
                " is later than the last trading minute of calendar"
                f" '{self.calendar.name}' ('{self.calendar.last_session}')."
            )
        else:
            msg += f" is not a trading minute of calendar '{self.calendar.name}'."
        return msg


class MinuteOutOfBounds(ValueError):
    """A minute required to be within bounds of trading minutes is not.

    Parameters
    ----------
    calendar
        Calendar for which `minute` required to be within bounds of
        trading minutes.

    minute
        Minute required to be within bounds of `calendar`'s trading
        minutes.

    param_name
        Name of a parameter receiving `minute`.
    """

    def __init__(
        self, calendar: ExchangeCalendar, minute: pd.Timestamp, param_name: str
    ):
        self.calendar = calendar
        self.minute = minute
        self.param_name = param_name

    def __str__(self) -> str:
        msg = f"Parameter `{self.param_name}` receieved as '{self.minute}' although"
        if self.minute < self.calendar.first_minute:
            msg += (
                " cannot be earlier than the first trading minute of calendar"
                f" '{self.calendar.name}' ('{self.calendar.first_minute}')."
            )
        elif self.minute > self.calendar.last_minute:
            msg += (
                " cannot be later than the last trading minute of calendar"
                f" '{self.calendar.name}' ('{self.calendar.last_minute}')."
            )
        else:
            assert (
                self.minute < self.calendar.first_minute
                or self.minute > self.calendar.last_minute
            )
        return msg


class RequestedSessionOutOfBounds(ValueError):
    """The requested session would fall beyond calendar bounds.

    Parameters
    ----------
    calendar
        Calendar for which session would be out-of-bounds.

    too_early
        True if requested session would be earlier than the first calendar
            session.
        False if requested session would be later than the last calendar
            session.
    """

    def __init__(self, calendar: ExchangeCalendar, too_early: bool):
        self.calendar = calendar
        self.adverb = "before" if too_early else "after"
        self.position = "first" if too_early else "last"
        self.bound = calendar.first_session if too_early else calendar.last_session

    def __str__(self) -> str:
        return (
            f"Requested session would fall {self.adverb} the calendar's {self.position}"
            f" session ('{self.bound}')."
        )


class RequestedMinuteOutOfBounds(ValueError):
    """The requested trading minute would fall beyond calendar bounds.

    Parameters
    ----------
    calendar
        Calendar for which minute would be out-of-bounds.

    too_early
        True if requested minute would be earlier than the first calendar
            minute.
        False if requested minute would be later than the last calendar
            minute.
    """

    def __init__(self, calendar: ExchangeCalendar, too_early: bool):
        self.calendar = calendar
        self.adverb = "before" if too_early else "after"
        self.position = "first" if too_early else "last"
        self.bound = calendar.first_minute if too_early else calendar.last_minute

    def __str__(self) -> str:
        return (
            f"Requested minute would fall {self.adverb} the calendar's {self.position}"
            f" trading minute ('{self.bound}')."
        )


class IndexOverlapError(ValueError):
    """Periods implied by indices overlap."""


class IntervalsOverlapError(IndexOverlapError):
    """Intervals of requested trading index would overlap."""

    # pylint: disable=missing-return-type-doc
    def __str__(self):  # noqa: D105
        return (
            "Unable to create trading index as intervals would overlap."
            " This can occur when the frequency is longer than a break or"
            " the gap between one session's close and the next session's"
            " open. To shorten intervals that would otherwise overlap"
            " either pass `curtail_overlaps` as True or pass"
            " `force_close` and/or `force_break_close` as True."
        )


class IndicesOverlapError(IndexOverlapError):
    """Indices of requested trading index would overlap."""

    # pylint: disable=missing-return-type-doc
    def __str__(self):  # noqa: D105
        return (
            "Unable to create trading index as an indice would fall to the"
            " right of (later than) the subsequent indice. This can occur"
            " when the frequency is longer than a break or the frequency"
            " is longer than the gap between one session's close and the"
            " next session's open. Consider  passing `closed` as `left`"
            " or passing `force_close` and/or `force_break_close`"
            " as True."
        )



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import Holiday
import pytz

from .common_holidays import (
    eid_al_adha_first_day,
    eid_al_fitr_first_day,
    european_labour_day,
    new_years_day,
)
from .exchange_calendar import WEEKDAYS, WEEKENDS, HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

NationalSovereigntyAndChildrensDay = Holiday(
    "National Sovereignty and Children's Day",
    month=4,
    day=23,
)

LabourDay = european_labour_day(start_date="2009")

CommemorationOfAttaturkYouthAndSportsDay = Holiday(
    "Commemoration of Attaturk, Youth and Sports Day",
    month=5,
    day=19,
)

EidAlFitr1 = eid_al_fitr_first_day

EidAlFitr2 = eid_al_fitr_first_day + timedelta(1)

EidAlFitr3 = eid_al_fitr_first_day + timedelta(2)

DemocracyAndNationalUnityDay = Holiday(
    "Democracy and National Unity Day", month=7, day=15, start_date="2017"
)

EidAlAdha1 = eid_al_adha_first_day

EidAlAdha2 = eid_al_adha_first_day + timedelta(1)

EidAlAdha3 = eid_al_adha_first_day + timedelta(2)

EidAlAdha4 = eid_al_adha_first_day + timedelta(3)

VictoryDay = Holiday(
    "Victory Day",
    month=8,
    day=30,
)

RepublicDay = Holiday(
    "Republic Day",
    month=10,
    day=29,
)

# Early Closes
DayBeforeEidAlFitr = eid_al_fitr_first_day - timedelta(1)

EidAlFitrHalfDay = [day for day in DayBeforeEidAlFitr if day.weekday() not in WEEKENDS]

DayBeforeEidAlAdha = eid_al_adha_first_day - timedelta(1)

EidAlAdhaHalfDay = [day for day in DayBeforeEidAlAdha if day.weekday() not in WEEKENDS]

RepublicDayHalfDay = Holiday(
    "Republic Day half day",
    month=10,
    day=28,
    days_of_week=WEEKDAYS,
)


class XISTExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Istanbul Stock Exchange (XIST).

    # Note: Effective as of Nov 14, 2016 according to official website
    Open Time: 10:00 AM, TRT
    Close Time: 6:00 PM, TRT

    Regularly-Observed Holidays:
    - New Year's Day
    - National Sovereignty and Children's Day (Apr 23)
    - Labour Day (May 1)
    - Commemoration of Attaturk, Youth and Sports Day (May 19)
    - Eid-ul-Fitr (Islamic Lunar Calendar)
    - Democracy and National Unity Day (Jul 15)
    - Eid-al-Adha (first day is a half-day, remaining full holidays)
    - Victory Day (Aug 30)
    - Republic Day (Oct 29)

    Early Closes:
    - First Day of Eid-al-Adha
    - Republic Day (Oct 28)
    """

    name = "XIST"

    tz = pytz.timezone("Europe/Istanbul")

    open_times = ((None, time(10)),)

    close_times = ((None, time(18, 00)),)

    regular_early_close = time(12, 30)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NationalSovereigntyAndChildrensDay,
                LabourDay,
                CommemorationOfAttaturkYouthAndSportsDay,
                DemocracyAndNationalUnityDay,
                VictoryDay,
                RepublicDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        misc_adhoc = [
            pd.Timestamp("2002-01-04"),  # Market Holiday
            pd.Timestamp("2003-02-14"),  # Eid al Adha extra holiday
            pd.Timestamp("2003-11-21"),  # Terror attacks
            pd.Timestamp("2003-11-24"),  # Eid al Fitr extra holiday
            pd.Timestamp("2003-11-28"),  # Eid al Fitr extra holiday
            pd.Timestamp("2004-01-23"),  # Bad weather
            pd.Timestamp("2004-12-30"),  # Closure for redenomination
            pd.Timestamp("2004-12-31"),  # Closure for redenomination
            pd.Timestamp("2006-01-13"),  # Eid al Adha extra holiday
        ]

        return list(
            chain(
                misc_adhoc,
                EidAlFitr1,
                EidAlFitr2,
                EidAlFitr3,
                EidAlAdha1,
                EidAlAdha2,
                EidAlAdha3,
                EidAlAdha4,
            )
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar([RepublicDayHalfDay]),
            )
        ]

    @property
    def special_closes_adhoc(self):
        early_close_days = pd.DatetimeIndex(EidAlFitrHalfDay + EidAlAdhaHalfDay)
        return [(self.regular_early_close, early_close_days)]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    nearest_workday,
    next_monday,
    previous_friday,
)
from pytz import timezone

from .common_holidays import european_labour_day, new_years_day
from .lunisolar_holidays import (
    chinese_lunar_new_year_dates,
    dragon_boat_festival_dates,
    mid_autumn_festival_dates,
    qingming_festival_dates,
)
from .exchange_calendar import (
    SATURDAY,
    SUNDAY,
    THURSDAY,
    TUESDAY,
    HolidayCalendar,
    ExchangeCalendar,
)


def before_chinese_new_year_offset(holidays):
    """
    For Holidays that come before Chinese New Year, we subtract a day
    and then move any weekends to previous friday.
    """
    return pd.to_datetime(holidays.map(lambda d: previous_friday(d)))


def chinese_new_year_offset(holidays):
    """
    For Holidays on or after Chinese New Year, we add a day
    and then move any weekends to next monday.
    """
    return pd.to_datetime(holidays.map(lambda d: next_monday(d)))


def nearest_workday_after_2013(dt):
    """
    Nearest workday starting in 2014.
    """
    return nearest_workday(dt) if dt.year > 2013 else dt


def manual_nearest_workday(holidays):
    """
    Nearest workday observance rule for Chinese lunar calendar holidays.
    The nearest workday rule seems to start in 2014 for these holidays.
    """
    return pd.to_datetime(holidays.map(lambda d: nearest_workday_after_2013(d)))


def manual_extra_days(holidays):
    """
    Four day weekend makeup days for Chinese lunar calendar holidays.
    The four day weekend rule seem to start in 2007 for these holidays.
    """
    friday_extras = [
        d + timedelta(1) for d in holidays if d.weekday() == THURSDAY and d.year > 2006
    ]

    monday_extras = [
        d - timedelta(1) for d in holidays if d.weekday() == TUESDAY and d.year > 2006
    ]

    return pd.to_datetime(friday_extras + monday_extras)


def taiwan_makeup_rule(holidays):
    """
    This function attempts to implement what seems to be the Taiwan holiday
    observance rule since 2013.

    Notes
    -----
    If a holiday falls on a Tuesday or Thursday, and extra holiday is observed
    on Monday/Friday to create a four day weekend.  If a holiday falls on a
    weekend, it is observed on the nearest workday (Monday/Friday).
    """
    # Create four day weekends
    mon = holidays[holidays.weekday == TUESDAY] - timedelta(1)
    fri = holidays[holidays.weekday == THURSDAY] + timedelta(1)

    # Nearest workday makeups
    mon_makeups = holidays[holidays.weekday == SUNDAY] + timedelta(1)
    fri_makeups = holidays[holidays.weekday == SATURDAY] - timedelta(1)

    # Only observe rule after 2013
    extras = pd.to_datetime(list(chain(mon, mon_makeups, fri, fri_makeups)))
    extras = extras[extras.year > 2013]

    return holidays.append(extras)


NewYearsDay = new_years_day(observance=taiwan_makeup_rule)

PeaceMemorialDay = Holiday(
    "Peace Memorial Day",
    month=2,
    day=28,
    observance=taiwan_makeup_rule,
)

WomenAndChildrensDay = Holiday(
    "Women and Children's Day",
    month=4,
    day=4,
    start_date="2011",
    observance=taiwan_makeup_rule,
)

LabourDay = european_labour_day(observance=manual_nearest_workday)

NationalDay = Holiday(
    "National Day of the Republic of China",
    month=10,
    day=10,
    observance=taiwan_makeup_rule,
)

chinese_new_year = chinese_new_year_offset(chinese_lunar_new_year_dates)

chinese_new_years_eve = before_chinese_new_year_offset(
    chinese_new_year - timedelta(1),
)

chinese_new_years_eve_2 = before_chinese_new_year_offset(
    chinese_new_years_eve - timedelta(1),
)

chinese_new_year_2 = chinese_new_year_offset(
    chinese_new_year + timedelta(1),
)

chinese_new_year_3 = chinese_new_year_offset(
    chinese_new_year_2 + timedelta(1),
)

tomb_sweeping_day = manual_nearest_workday(qingming_festival_dates)

tomb_sweeping_day_extras = manual_extra_days(tomb_sweeping_day)

dragon_boat_festival = manual_nearest_workday(dragon_boat_festival_dates)

dragon_boat_festival_extras = manual_extra_days(dragon_boat_festival)

mid_autumn_festival = manual_nearest_workday(
    mid_autumn_festival_dates,
)

mid_autumn_festival_extras = manual_extra_days(mid_autumn_festival)

# Taiwan takes multiple days off before and after chinese new year,
# and sometimes it is unclear precisely which days will be holidays.
chinese_new_year_extras = pd.to_datetime(
    [
        "2002-02-07",
        "2002-02-15",
        "2003-01-29",
        "2004-01-19",
        "2005-02-04",
        "2006-02-02",
        "2007-02-22",
        "2007-02-23",
        "2008-02-04",
        "2009-01-29",
        "2009-01-30",
        "2010-02-18",
        "2010-02-19",
        "2011-01-31",
        "2012-01-26",
        "2012-01-27",
        "2013-02-14",
        "2013-02-15",
        "2014-01-28",
        "2015-02-16",
        "2016-02-11",
        "2016-02-12",
        "2017-01-25",
        "2018-02-13",
        "2019-01-31",
        "2019-02-08",
        "2020-01-21",
        "2020-01-22",
        "2021-02-08",
        "2021-02-09",
        "2022-01-27",
    ]
)

# Some abnormal observances of regularly observed holidays.
extra_holidays = pd.to_datetime(
    [
        "2021-04-02",  # Women And Childrens Day
        "2020-04-02",  # Tomb Sweeping Day
        "2016-04-05",  # Tomb Sweeping Day
        "2012-12-31",  # New Year's Eve
        "2012-02-27",  # Peace Memorial Day
        "2009-01-02",  # New Year's Day
        "2006-10-09",  # National Day
        "2005-09-01",  # Bank Holiday
    ]
)

typhoons = pd.to_datetime(
    [
        "2019-09-30",
        "2019-08-09",
        "2016-09-28",
        "2016-09-27",
        "2016-07-08",
        "2015-09-29",
        "2015-07-10",
        "2014-07-23",
        "2013-08-21",
        "2012-08-02",
        "2009-08-07",
        "2008-09-29",
        "2008-07-28",
        "2007-09-18",
        "2005-08-05",
        "2005-07-18",
        "2004-10-25",
        "2004-08-25",
        "2004-08-24",
        "2002-09-06",
    ]
)


class XTAIExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Taiwan Stock Exchange Corporation (XTAI).

    Open Time: 9:00 AM, CST
    Close Time: 1:30 PM, CST

    Regularly-Observed Holidays:
    - New Year's Day
    - Chinese New Year's Eve
    - Chinese New Year
    - Peace Memorial Day (Feb 28)
    - Women and Children's Day (Apr 4)
    - Tomb Sweeping Day (Lunar Calendar)
    - Labour Day (May 1)
    - Dragon Boat Festival (Lunar Calendar)
    - Mid-Autumn Festival (Lunar Calendar)
    - National Day of the Republic of China (Oct 10)


    Early Closes:
    - None
    """

    name = "XTAI"

    tz = timezone("Asia/Taipei")

    open_times = ((None, time(9)),)

    close_times = ((None, time(13, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                PeaceMemorialDay,
                WomenAndChildrensDay,
                LabourDay,
                NationalDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                extra_holidays,
                typhoons,
                chinese_new_years_eve,
                chinese_new_years_eve_2,
                chinese_new_year,
                chinese_new_year_2,
                chinese_new_year_3,
                chinese_new_year_extras,
                tomb_sweeping_day,
                tomb_sweeping_day_extras,
                dragon_boat_festival,
                dragon_boat_festival_extras,
                mid_autumn_festival,
                mid_autumn_festival_extras,
            )
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Useful source for the history of New Zealand public holidays:
# Nancy Swarbrick, 'Public holidays', Te Ara - the Encyclopedia of New Zealand
# http://www.TeAra.govt.nz/en/public-holidays/print (accessed 22 December 2019)

from datetime import time

from pandas import Timestamp
from pandas.tseries.holiday import (
    MO,
    DateOffset,
    EasterMonday,
    GoodFriday,
    Holiday,
    next_monday,
    next_monday_or_tuesday,
    previous_workday,
    weekend_to_monday,
)
from pytz import timezone

from .common_holidays import anzac_day, boxing_day, christmas, new_years_day
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# Prior to 2015, Waitangi Day and Anzac Day are not "Mondayized",
# that is, if they occur on the weekend, there is no make-up.
MONDAYIZATION_START_DATE = "2015-01-01"

# Regular Holidays
# ----------------
NewYearsDay = new_years_day(observance=next_monday)

DayAfterNewYearsDay = Holiday(
    "Day after New Year's Day",
    month=1,
    day=2,
    observance=next_monday_or_tuesday,
)

WaitangiDayNonMondayized = Holiday(
    "Waitangi Day",
    month=2,
    day=6,
    end_date=MONDAYIZATION_START_DATE,
)

WaitangiDay = Holiday(
    "Waitangi Day",
    month=2,
    day=6,
    observance=weekend_to_monday,
    start_date=MONDAYIZATION_START_DATE,
)

AnzacDayNonMondayized = anzac_day(end_date=MONDAYIZATION_START_DATE)

AnzacDay = anzac_day(
    observance=weekend_to_monday,
    start_date=MONDAYIZATION_START_DATE,
)

QueensBirthday = Holiday(
    "Queen's Birthday",
    month=6,
    day=1,
    offset=DateOffset(weekday=MO(1)),
)

LabourDay = Holiday(
    "Labour Day",
    month=10,
    day=1,
    offset=DateOffset(weekday=MO(4)),
)

Christmas = christmas(observance=next_monday)

BoxingDay = boxing_day(observance=next_monday_or_tuesday)


# Early Closes
# ------------
BusinessDayPriorToChristmasDay = Holiday(
    "Business Day prior to Christmas Day",
    month=12,
    day=25,
    observance=previous_workday,
    start_date="2011-01-01",
)

BusinessDayPriorToNewYearsDay = Holiday(
    "Business Day prior to New Year's Day",
    month=1,
    day=1,
    observance=previous_workday,
    start_date="2011-01-01",
)

# Further ad-hoc holidays and closures sourced from gaps in index series
XMAS22DEC86 = Timestamp("1986-12-22")
XMAS23DEC86 = Timestamp("1986-12-23")
XMAS24DEC86 = Timestamp("1986-12-24")
XMAS29DEC86 = Timestamp("1986-12-29")
XMAS30DEC86 = Timestamp("1986-12-30")
XMAS31DEC86 = Timestamp("1986-12-31")
XMAS24DEC90 = Timestamp("1990-12-24")
XMAS31DEC90 = Timestamp("1990-12-31")
XMAS27DEC91 = Timestamp("1991-12-27")
NYDX03JAN92 = Timestamp("1992-01-03")
XMAS31DEC99 = Timestamp("1999-12-31")


class XNZEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the New Zealand Exchange (NZX).

    Open Time: 10:00 AM, NZ
    Close Time: 4:45 PM, NZ

    Regularly-Observed Holidays:
    - New Year's Day
    - Day after New Year's Day
    - Waitangi Day
    - Good Friday
    - Easter Monday
    - Anzac Day
    - Queen's Birthday
    - Labour Day
    - Christmas
    - Boxing Day

    Early Closes:
    - Business Day prior to Christmas Day
    - Business Day prior to New Year's Day
    """

    regular_early_close = time(12, 45)

    name = "XNZE"

    tz = timezone("NZ")

    open_times = ((None, time(10)),)

    close_times = ((None, time(16, 45)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                DayAfterNewYearsDay,
                WaitangiDayNonMondayized,
                WaitangiDay,
                GoodFriday,
                EasterMonday,
                AnzacDayNonMondayized,
                AnzacDay,
                QueensBirthday,
                LabourDay,
                Christmas,
                BoxingDay,
            ]
        )

    # add extra ad-hoc holidays
    @property
    def adhoc_holidays(self):
        return [
            XMAS22DEC86,
            XMAS23DEC86,
            XMAS24DEC86,
            XMAS29DEC86,
            XMAS30DEC86,
            XMAS31DEC86,
            XMAS24DEC90,
            XMAS31DEC90,
            XMAS27DEC91,
            NYDX03JAN92,
            XMAS31DEC99,
        ]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [BusinessDayPriorToChristmasDay, BusinessDayPriorToNewYearsDay]
                ),
            )
        ]
